#
# Copyright 2007 Fedora Unity
#
# Jonathan Steffan <jon a fedoraunity.org>
# Jeroen van Meeuwen <kanarip a fedoraunity.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 only
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import revisor
import yum
from ConfigParser import SafeConfigParser
from ConfigParser import RawConfigParser
import string
import subprocess
import time
import math
import shutil
import re
import fnmatch
import yum.misc
import yum.packageSack

import revisor.cfg
import revisor.progress
import revisor.kickstart as kickstart
from revisor import lm_optical

# Import constants
from revisor.constants import *

# Translation
from rhpl.translate import _, N_, getDefaultLangs
import rhpl.translate as translate

class RevisorBase:
    """RevisorBase holds common functions shared amongst our CLI and GUI mode"""
    def __init__(self, cfg = None, log = None):

        if cfg:
            self.cfg = cfg
        if log:
            self.log = log

        self.cfg.setup_yum()
        self.cfg.setup_ks()

    def show_help(self, keyword):
        if self.cfg.revisorUseGUI:
            self.mode.base_buttons_xml.get_widget("button_information").set_sensitive(False)
            pid = os.fork()
            if not pid:
                self.log.info(_("Opening up /usr/bin/htmlview %s%s") % (DOCS_BASEPATH, keyword))
                os.execv("/usr/bin/htmlview", ["/usr/bin/htmlview", "%s%s" % (DOCS_BASEPATH, keyword)])
            else:
#                self.mode.base_buttons_xml.get_widget("button_information").set_sensitive(True)
                self.log.error(_("Cannot fork process showing help, please use %s%s") % (DOCS_BASEPATH, keyword))
        else:
            self.log.warning(_("Cannot show Help in CLI mode, use %s%s") % (DOCS_BASEPATH, keyword))

    def setup_yum(self):
        """Setup the yum object"""
        self.cfg.yumobj.pbar = self.progress_bar(_("Loading Repositories"), callback=True)

        try:
            if hasattr(self.cfg.yumobj,"_getConfig"):
                self.cfg.yumobj._getConfig(fn=self.cfg.main_conf, plugin_types=(yum.plugins.TYPE_CORE,))
            else:
                self.cfg.yumobj.doConfigSetup(fn=self.cfg.main_conf, plugin_types=(yum.plugins.TYPE_CORE,))
                self.log.debug(_("Using deprecated YUM function: %s()") % "doConfigSetup")
        except:
            self.log.error(_("yum.YumBase.doConfigSetup failed, probably an invalid configuration file %s") % self.cfg.main_conf, recoverable=False)

        try:
            if not self.cfg.repo_override:
                self.cfg.reposSetup(callback = self.cfg.yumobj.pbar)
                repo_setup = True
            if self.cfg.repo_override:
                # Clear all the repos we have enabled from the model, and inject our override repo(s)
                for repo in self.cfg.yumobj.repos.repos.keys():
                    self.cfg.yumobj.repos.delete(str(repo))

                for repo in self.cfg.repo_override:
                    repoObj = yum.yumRepo.YumRepository(repo["name"])
                    repoObj.setAttribute('name', repo["name"])
                    repoObj.setAttribute('baseurl', repo["baseurl"])
                    repoObj.setAttribute('gpgcheck', repo["gpgcheck"])
                    self.cfg.yumobj.repos.add(repoObj)
                    self.cfg.yumobj.repos.enableRepo(repoObj.id)

                repo_setup=True

        except yum.Errors.LockError:
            self.cfg.yumobj.pbar.destroy()
            self.log.error(_("Another application is running which is accessing software information."), recoverable = False)
            repo_setup = False

        except revisor.errors.RevisorDownloadError, e:
            self.cfg.yumobj.pbar.destroy()
            self.log.error(_("Unable to retrieve software information. This could be caused by not having a network connection available."))
            repo_setup = False

        if repo_setup:
            if hasattr(self.cfg.yumobj,"_getTs"):
                self.cfg.yumobj._getTs()
            else:
                self.log.debug(_("Using deprecated YUM function: %s()") % "doTsSetup")
                self.cfg.yumobj.doTsSetup()
            if hasattr(self.cfg.yumobj,"_getRpmDB"):
                self.cfg.yumobj._getRpmDB()
            elif hasattr(self.cfg.yumobj,"doSetupRPMDB"):
                self.log.debug(_("Using deprecated YUM function: %s()") % "doSetupRPMDB")
                self.cfg.yumobj.doSetupRPMDB()
            elif hasattr(self.cfg.yumobj,"doRpmDBSetup"):
                self.log.debug(_("Apparently we're running on an Enterprise Linux system (we can tell from the way yum sets up the RPM Database)"))
                self.cfg.yumobj.doRpmDBSetup()

            if hasattr(self.cfg.yumobj,"_getRepos"):
                self.cfg.yumobj._getRepos()
            else:
                self.log.debug(_("Using deprecated YUM function: %s()") % "doRepoSetup")
                self.cfg.yumobj.doRepoSetup()

#            self.log.debug(_("Arch list = %s") % self.cfg.arch_list)

            try:
                if hasattr(self.cfg.yumobj,"_getSacks"):
                    self.cfg.yumobj._getSacks(archlist=self.cfg.arch_list)
                else:
                    self.log.debug(_("Using deprecated YUM function: %s()") % "doSackSetup")
                    self.cfg.yumobj.doSackSetup(archlist=self.cfg.arch_list)
            except:
                pass

            self.cfg.yumobj.pbar.destroy()
            return True
        else:
            self.cfg.yumobj.pbar.destroy()
            return False

    def pkglist_from_ksdata(self, groupList=[], packageList=[], excludedList=[], ignore_list=[]):
        """Takes valid groupList, packageList and excludedList such as in kickstart packages manifests,
        and selects packages from these lists. The way it does so depends on whether we are doing
        Installation Media Respins"""

        self.log.debug(_("Building a nice package list from ksdata, and adding it to the transaction"))

        # Actually, let's figure out what mode we're in and call something once we know which way to go

        if self.cfg.mode_respin and self.cfg.media_installation:
            self.pkglist_from_ksdata_respin(groupList, packageList, excludedList)
        else:
            self.pkglist_from_ksdata_normal(groupList, packageList, excludedList, ignore_list)

    def pkglist_from_ksdata_normal(self, groupList, packageList, excludedList, ignore_list=[]):
        """This function adds packages to the yum transaction set the regular way
        Pass it a list of names of packages that we should ignore the architecture off
        An example is libbeagle, kernel, pm-utils"""

        total = float(len(groupList) * 3 +
                    len(packageList) * 2 +
                    len(excludedList))
        current = 0.0

        pbar = self.progress_bar(_("Select kickstart packages"))

##
## Add packages from Groups in ksdata
##
        for ksgroup in groupList:
            current += 3.0
            self.log.debug(_("Found group: %s") % ksgroup.name)
            # Get group object from comps
            grp = self.cfg.yumobj.comps.return_group(ksgroup.name)

            # Select the group
            txmbrs_used = self.cfg.yumobj.selectGroup(grp.name)

            # Add all required packages
            if ksgroup.include >= kickstart.constants.GROUP_REQUIRED:
                self.log.debug(_("Selecting required packages for group %s") % grp.name)
                for pkg in grp.mandatory_packages.keys():
                    if pkg in excludedList:
                        continue
                    self.log.debug(_("Including %s") % pkg)
                    # Get the package objects
                    try:
                        pkgs = self.cfg.yumobj.pkgSack.returnNewestByName(pkg)
                        if len(pkgs) > 1 and not pkg in ignore_list:
                            pkgs = self.cfg.yumobj.bestPackagesFromList(pkgs)
                        # Add to transaction
                        for po in pkgs:
                            self.cfg.yumobj.tsInfo.addInstall(po)
                            self.log.debug(_("From Groups --optional: Adding %s-%s:%s-%s.%s to transaction") % (po.name, po.epoch, po.version, po.release, po.arch))
                    except yum.Errors.PackageSackError, e:
                        if self.cfg.ksobj._get("packages","handleMissing") != kickstart.constants.KS_MISSING_IGNORE:
                            self.log.warning(e.value)

            # Add all default packages
            if ksgroup.include >= kickstart.constants.GROUP_DEFAULT:
                self.log.debug(_("Selecting default packages for group %s") % grp.name)
                for pkg in grp.default_packages.keys():
                    if pkg in excludedList:
                        continue
                    self.log.debug(_("Including %s") % pkg)
                    # Get the package objects
                    try:
                        pkgs = self.cfg.yumobj.pkgSack.returnNewestByName(pkg)
                        if len(pkgs) > 1 and not pkg in ignore_list:
                            pkgs = self.cfg.yumobj.bestPackagesFromList(pkgs)
                        # Add to transaction
                        for po in pkgs:
                            self.cfg.yumobj.tsInfo.addInstall(po)
                            self.log.debug(_("From Groups --optional: Adding %s-%s:%s-%s.%s to transaction") % (po.name, po.epoch, po.version, po.release, po.arch))
                    except yum.Errors.PackageSackError, e:
                        if self.cfg.ksobj._get("packages","handleMissing") != kickstart.constants.KS_MISSING_IGNORE:
                            self.log.warning(e.value)

            # Add all optional packages
            if ksgroup.include >= kickstart.constants.GROUP_ALL:
                self.log.debug(_("Selecting optional packages for group %s") % grp.name)
                # All optional packages included, get the package names
                for pkg in grp.optional_packages.keys():
                    if pkg in excludedList:
                        continue
                    self.log.debug(_("Including %s") % pkg)
                    # Get the package objects
                    try:
                        pkgs = self.cfg.yumobj.pkgSack.returnNewestByName(pkg)
                        if len(pkgs) > 1 and not pkg in self.cfg.archs_pkgs_installation:
                            pkgs = self.cfg.yumobj.bestPackagesFromList(pkgs)
                        # Add to transaction
                        for po in pkgs:
                            if not po.name in ignore_list:
                                self.cfg.yumobj.tsInfo.addInstall(po)
                                self.log.debug(_("From Groups --optional: Adding %s-%s:%s-%s.%s to transaction") % (po.name, po.epoch, po.version, po.release, po.arch))
                            else:
                                self.log.debug(_("From Groups --optional: Not adding %s now because it is in the ignore list") % po.name)
                    except yum.Errors.PackageSackError, e:
                        if self.cfg.ksobj._get("packages","handleMissing") != kickstart.constants.KS_MISSING_IGNORE:
                            self.log.warning(e.value)
            pbar.set_fraction(current/total)

##
## Pull the conditional packages in
##
            for condreq, cond in grp.conditional_packages.iteritems():
                self.log.debug(_("Testing condition: %s / %s") % (condreq, cond))
                pkgs = self.cfg.yumobj.pkgSack.searchNevra(name=condreq)
                if len(pkgs) > 1:
                    pkgs = self.cfg.yumobj.bestPackagesFromList(pkgs)
                if self.cfg.yumobj.tsInfo.conditionals.has_key(cond):
                    self.cfg.yumobj.tsInfo.conditionals[cond].extend(pkgs)
                else:
                    self.cfg.yumobj.tsInfo.conditionals[cond] = pkgs
##
## Add packages in ksdata to transaction
##
        for pkg in packageList:
            current += 2.0
            self.log.debug(_("From package list, including: %s") % pkg)
            if pkg in excludedList:
                continue
            try:
                pkgs = self.cfg.yumobj.pkgSack.returnNewestByName(pkg)
                if len(pkgs) > 1 and not pkg in ignore_list:
                    pkgs = self.cfg.yumobj.bestPackagesFromList(pkgs)

                for po in pkgs:
                    self.cfg.yumobj.tsInfo.addInstall(po)
                    self.log.debug(_("From Packages: Adding %s-%s:%s-%s.%s to transaction") % (po.name, po.epoch, po.version, po.release, po.arch))
            except yum.Errors.PackageSackError, e:
                # Let's see if there's some regexp, and if there is, let's search
                if "*" in pkg:
                    try:
                        pkglist = self.cfg.yumobj.pkgSack.simplePkgList()
                        matches = []
                        # anything we couldn't find a match for
                        # could mean it's not there, could mean it's a wildcard
                        if re.match('.*[\*,\[,\],\{,\},\?].*', pkg):
                            restring = fnmatch.translate(pkg)
                            regex = re.compile(restring, flags=re.I) # case insensitive
                            for item in pkglist:
                                if regex.match(item[0]):
                                    matches.append(item[0])
                                    self.log.debug(_("Found packages matching '%s': %s") % (pkg,item[0]))
                        for pkg in matches:
                            pkgs = self.cfg.yumobj.pkgSack.returnNewestByName(pkg)
                            if len(pkgs) > 1:
                                pkgs = self.cfg.yumobj.bestPackagesFromList(pkgs)

                            for po in pkgs:
                                if not po.name in ignore_list:
                                    self.cfg.yumobj.tsInfo.addInstall(po)
                                    self.log.debug(_("From Packages: Adding %s-%s:%s-%s.%s to transaction") % (po.name, po.epoch, po.version, po.release, po.arch))
                                else:
                                    self.log.debug(_("From Packages: Not adding %s now because it is in the ignore list") % po.name)

                    except yum.Errors.PackageSackError, e:
                        if self.cfg.ksobj._get("packages","handleMissing") != kickstart.constants.KS_MISSING_IGNORE:
                            self.log.warning(e.value)
                else:
                    if self.cfg.ksobj._get("packages","handleMissing") != kickstart.constants.KS_MISSING_IGNORE:
                        self.log.warning(e.value)
            pbar.set_fraction(current/total)

##
## Exclude packages from ksdata
##
        for pkg in excludedList:
            current += 1.0
            try:
                pkgs = self.cfg.yumobj.pkgSack.returnNewestByName(pkg)
                for po in pkgs:
                    self.log.debug(_("From Excludes: Removing %s-%s:%s-%s.%s from transaction") % (po.name, po.epoch, po.version, po.release, po.arch))
                    self.cfg.yumobj.tsInfo.remove(po.pkgtup)
# FIXME
# Bwegh this logs warnings we don't want that
#            except yum.Errors.PackageSackError, e:
#                self.log.warning(e.value)
            except:
                pass
            pbar.set_fraction(current/total)

        self.cfg.yumobj.tsInfo.makelists()

        self.log.debug(_("This is what was selected to be installed:"))
        installpkgs = []
        for po in self.cfg.yumobj.tsInfo.installed:
            installpkgs.append(po.pkgtup)
            self.log.debug("--> %s" % str(po.pkgtup))

        pbar.destroy()

    def pkglist_from_ksdata_respin(self):
        """Takes valid kshandler.packages.groupList, kshandler.packages.packageList
        and kshandler.packages.excludedList, and builds a nice and simple list that
        we can add our required packages to, and then start depsolving"""

        pkglist = []
        matchdict = {}
##
## Add packages from Groups in ksdata
##
        for ksgroup in groupList:
            if not self.cfg.yumobj.comps.has_group(ksgroup.name):
                self.log.warning(_("No such group %s") % ksgroup.name )
                continue
            else:
                self.log.debug(_("Found group: %s") % ksgroup.name)

            # Get group object from comps
            grp = self.cfg.yumobj.comps.return_group(ksgroup.name)
            txmbrs_used = self.cfg.yumobj.selectGroup(grp.name)

            if ksgroup.include >= kickstart.constants.GROUP_REQUIRED:
                pkglist.extend(grp.mandatory_packages.keys())
            if ksgroup.include >= kickstart.constants.GROUP_DEFAULT:
                pkglist.extend(grp.default_packages.keys())
            if ksgroup.include >= kickstart.constants.GROUP_ALL:
                pkglist.extend(grp.optional_packages.keys())

##
## Pull the conditional packages in
##
            for condreq, cond in grp.conditional_packages.iteritems():
                self.log.debug(_("Testing condition: %s / %s") % (condreq, cond))
                pkgs = self.cfg.yumobj.pkgSack.searchNevra(name=condreq)
                if len(pkgs) > 1:
                    pkgs = self.cfg.yumobj.bestPackagesFromList(pkgs)
                if self.cfg.yumobj.tsInfo.conditionals.has_key(cond):
                    self.cfg.yumobj.tsInfo.conditionals[cond].extend(pkgs)
                else:
                    self.cfg.yumobj.tsInfo.conditionals[cond] = pkgs

##
## Add packages in ksdata to transaction
##
        for pkg in packageList:
            pkglist.append(pkg)

##
## Exclude packages from ksdata
##
        for pkg in excludedList:
            if pkg in pkglist:
                self.cfg.yumobj.conf.exclude.extend(pkg)

##
## Add pkglist to install
##
        # Make the search list unique
        pkglist = yum.misc.unique(pkglist)

        # Search repos for things in our searchlist, supports globs
        (exactmatched, matched, unmatched) = yum.packages.parsePackages(self.cfg.yumobj.pkgSack.returnPackages(), pkglist, casematch=1)
        matches = exactmatched + matched

        # Populate a dict of package objects to their names
        for match in matches:
            matchdict[match.name] = match

        # Get the newest results from the search
        mysack = yum.packageSack.ListPackageSack(matches)
        for match in mysack.returnNewestByNameArch():
            self.cfg.yumobj.tsInfo.addInstall(match)
            self.log.debug(_("Adding %s-%s:%s-%s.%s") % (match.name, match.epoch, match.version, match.release, match.arch))

    def progress_bar(self, title = "", parent = None, xml = None, callback = False):
        """This function should be used to determine the type of progress bar we need.
        There's three types:
            - GUI Dialog Progress Bar (separate dialog, window pops up)
            - GUI Nested Progress Bar (no separate dialog, progress bar is in gui.frame_xml)
            - CLI Progress Bar

        This function also determines whether we should have a Callback Progress Bar"""

        self.log.debug(_("Initting progress bar for ") + title)

        if self.cfg.revisorUseGUI:
            try:
                if not self.mode.BuildMedia == None:
                    self.mode.BuildMedia.show_task_list()
            except:
                self.log.debug(_("Apparently we have not yet entered the Build Media stage"))

            if callback:
                if not self.mode.frame_xml.get_widget("part_progress") == None:
                    return revisor.progress.ProgressCallbackGUI(title = title, parent = self.mode, xml = self.mode.frame_xml)
                elif not self.mode.main_window == None:
                    return revisor.progress.ProgressCallbackGUI(title = title, parent = self.mode, xml = self.mode.frame_xml)
                else:
                    return revisor.progress.ProgressCallbackGUI(title = title, parent = parent, xml = xml)
            else:
                if not self.mode.frame_xml.get_widget("part_progress") == None:
                    return revisor.progress.ProgressGUI(title = title, parent = self.mode, xml = self.mode.frame_xml)
                elif not self.mode.main_window == None:
                    return revisor.progress.ProgressGUI(title = title, parent = self.mode, xml = xml)
                else:
                    return revisor.progress.ProgressGUI(title = title, parent = parent, xml = xml)
        else:
            if callback:
                return revisor.progress.ProgressCallbackCLI(title = title)
            else:
                return revisor.progress.ProgressCLI(title = title)

    def set_mode(self, mode):
        self.mode = mode

    def get_package_deps(self, po, pbar):
        """Add the dependencies for a given package to the
           transaction info"""

        self.log.debug(_("Checking dependencies for %s.%s") % (po.name, po.arch))

        pbar.cur_task += 1.0
        reqs = po.requires
        provs = po.provides

        for req in reqs:
            if self.resolved_deps.has_key(req):
                pbar.cur_task += 1.0
                pbar.set_fraction(pbar.cur_task/pbar.num_tasks)
                continue

            (r,f,v) = req
            if r.startswith('rpmlib(') or r.startswith('config('):
                pbar.cur_task += 1.0
                pbar.set_fraction(pbar.cur_task/pbar.num_tasks)
                continue

            deps = self.cfg.yumobj.whatProvides(r, f, v).returnPackages()

            if not deps:
#                pbar.cur_task += 1.0
                self.log.warning(_("Unresolvable dependency %s %s %s in %s.%s") % (r, f, v, po.name, po.arch))
                continue

            depsack = yum.packageSack.ListPackageSack(deps)

            for dep in depsack.returnNewestByNameArch():
#                pbar.cur_task += 1.0
#                for dep_req in dep.requires:
#                    if not self.resolved_deps.has_key(dep_req):
#                        pbar.num_tasks += 1.0
#                    else:
#                        pbar.cur_task += 1.0

                self.cfg.yumobj.tsInfo.addInstall(dep)

                self.log.debug(_("Added %s-%s:%s-%s.%s for %s-%s:%s-%s.%s (requiring %s %s %s)") % (dep.name, dep.epoch, dep.version, dep.release, dep.arch, po.name, po.epoch, po.version, po.release, po.arch, r, f, v))

    def check_dependencies_conflicts(self):
        self.log.debug(_("Checking dependencies - allowing conflicts within the package set"))
        pbar = self.progress_bar(_("Resolving Dependencies"), callback = True)
        pbar.num_tasks = float(len(self.cfg.yumobj.tsInfo.getMembers()))
        reqs = []
        for txmbr in self.cfg.yumobj.tsInfo.getMembers():
            reqs.append(txmbr.po.requires)
        reqs = yum.misc.unique(reqs)
        pbar.num_tasks += float(len(reqs))

        pbar.cur_task = 0.0
        self.resolved_deps = {}
        final_pkgobjs = {}

        moretoprocess = True
        while moretoprocess: # Our fun loop
            moretoprocess = False
            for txmbr in self.cfg.yumobj.tsInfo.getMembers():
                if not final_pkgobjs.has_key(txmbr.po):
                    final_pkgobjs[txmbr.po] = None # Add the pkg to our final list
                    self.get_package_deps(txmbr.po, pbar) # Get the deps of our package
                    moretoprocess = True
        pbar.destroy()

    def check_dependencies_no_conflicts(self):
        """Check the dependencies"""
        self.log.debug(_("Checking dependencies - not allowing any conflicts within the package set"))
        pbar = self.progress_bar(_("Resolving Dependencies"), callback = True)

        dsCB = revisor.progress.dscb(pbar, self.cfg.yumobj, self.cfg)
        self.cfg.yumobj.dsCallback = dsCB

        try:
            (result, msgs) = self.cfg.yumobj.buildTransaction()
        except yum.Errors.RepoError, errmsg:
            self.cfg.yumobj.dsCallback = None
            pbar.destroy()

            self.log.error(_("""Errors where encountered while downloading package headers:

%s""" % (errmsg,)))

        self.cfg.yumobj.dsCallback = None # we only wanted this here.  blah.
        pbar.destroy()

        if result == 1:

            self.log.error(_("""Unable to resolve dependencies for some packages selected:

%s""" % (string.join(msgs, "\n"))))

        self.cfg.yumobj.tsInfo.makelists()

    def check_dependencies(self, style=None):
        if not style:
            if self.cfg.dependency_resolve_allow_conflicts:
                self.check_dependencies_conflicts()
            else:
                self.check_dependencies_no_conflicts()
        else:
            style()

        self.populate_stats()

    def enable_source_repositories(self):
        # Reset the existing sacks
        self.cfg.yumobj._pkgSack = None

        archlist = self.cfg.arch_list + ['src']
        for repo in self.cfg.yumobj.repos.listEnabled():
            srcrepo = '%s-source' % repo.id
            if len(self.cfg.yumobj.repos.findRepos(srcrepo)) < 1:
                self.log.error(_("No such repository: %s") % srcrepo)
            for r in self.cfg.yumobj.repos.findRepos(srcrepo):
                self.log.debug(_("Enabling %s repository") % r.id)
                r.enable()
                # Setup the repo
                self.cfg.yumobj._getRepos(thisrepo=r.id,doSetup=True)
                # Setup pkgSack with 'src' in the archlist
                try:
                    self.cfg.yumobj._getSacks(archlist=archlist,thisrepo=r.id)
                except yum.Errors.RepoError, e:
                    self.log.error(e)

    def download_source_packages(self):
        # Make sure all source repositories are enabled
        self.enable_source_repositories()

        for po in self.polist:
            srpm = po.sourcerpm.split('.src.rpm')[0]
            if not srpm in self.srpmlist:
                self.srpmlist.append(srpm)

        for srpm in self.srpmlist:
            (sname, sver, srel) = srpm.rsplit('-', 2)
            try:
                srpmpos = self.cfg.yumobj.pkgSack.searchNevra(name=sname, ver=sver, rel=srel)
                for srpmpo in srpmpos:
                    if "source" in srpmpo.repoid:
                        if not srpmpo in self.srpmpolist:
                            self.srpmpolist.append(srpmpo)
            except IndexError:
                self.log.error(_("Error: Cannot find a source rpm for %s") % srpm)

#        print "%d SRPM POs of %d wanted SRPMs found for %d RPMs" % (len(self.srpmpolist),len(self.srpmlist),len(self.polist))

        pbar = self.progress_bar(_("Downloading Source Packages"))
        total = len(self.srpmpolist)
#        print "%d source packages to be downloaded" % total
        current = 0.0
        pbar.set_fraction(0.0)
        self.prev_download = ""
        self.attempt = 1
        self.current = 1

        for pkg in self.srpmpolist:
            n,a,e,v,r = pkg.pkgtup
            packages =  self.cfg.yumobj.pkgSack.searchNevra(n,e,v,r,a)
            for download in packages:
                repo = self.cfg.yumobj.repos.getRepo(download.repoid)
                if hasattr(download,"returnSimple"):
                    remote = download.returnSimple('relativepath')
                else:
                    remote = download.relativepath
                if os.path.exists(download.localPkg()) and os.path.getsize(download.localPkg()) == int(download.returnSimple('packagesize')):
                    continue
                path = repo.getPackage(download)
                if not os.path.exists(download.localPkg()) and os.path.exists(path):
                    try:
                        os.link(path,download.localPkg())
                    except:
                        if self.cfg.copy_local:
                            shutil.copy2(path,download.localPkg())
                        else:
                            os.symlink(path,download.localPkg())

                txt = _("Downloading %s") % remote
                if self.prev_download == "":
                    self.prev_download = download
                elif self.prev_download == download:
                    self.attempt += 1
                    self.current -= 1
                else:
                    self.attempt = 1
                    self.prev_download = download

                if self.attempt > 1:
                    txt = "#%d: %s" % (self.attempt,txt)
                self.log.debug(txt)
            current += 1
            pbar.set_fraction(current/total)

    def download_packages(self):
        dlpkgs = map(lambda x: x.po, filter(lambda txmbr:
                                            txmbr.ts_state in ("i", "u"),
                                            self.cfg.yumobj.tsInfo.getMembers()))

        self.polist = dlpkgs

        pbar = self.progress_bar(_("Downloading Packages"))

        dlCb = revisor.progress.dlcb(pbar, dlpkgs, log = self.log, cfg = self.cfg)
        self.cfg.yumobj.repos.setProgressBar(dlCb)
        try:
            probs = self.cfg.yumobj.downloadPkgs(dlpkgs, dlCb)
        except yum.Errors.RepoError, errmsg:
            self.log.error(errmsg)
        except IndexError:
            self.log.error(_("Unable to find a suitable mirror."))

        self.cfg.yumobj.repos.setProgressBar(None)

        if len(probs.keys()) > 0:
            errstr = []
            for key in probs.keys():
                errors = yum.misc.unique(probs[key])
                for error in errors:
                    errstr.append("%s: %s" %(key, error))

            details_str = string.join(errstr, "\n")
            self.log.error(_("Errors were encountered while downloading packages: %s") % details_str)

        if self.cfg.getsource:
            self.download_source_packages()

    def pkglist_required(self):

        packages_to_add = self.cfg.required_pkgs
        packages_to_skip = []
        for pkg_tuple in self.cfg.yumobj.tsInfo.pkgdict.keys():
            # Make sure we don't delete any user selected options with "best" package magic
            for package in packages_to_add:
                if str(pkg_tuple[0]) == str(package):
                    self.log.debug(_("Overriding auto package selection with user package selection for %s..." % package))
                    packages_to_skip.append(package)

        for pkg in self.cfg.required_pkgs:
            try:
                pkgs = self.cfg.yumobj.pkgSack.returnNewestByName(pkg)
                if not pkg in packages_to_skip:
                    pkgs = self.cfg.yumobj.bestPackagesFromList(pkgs)
                for po in pkgs:
                    self.cfg.yumobj.tsInfo.addInstall(po)
                    self.log.debug(_("Adding req. pkg %s-%s:%s-%s.%s") % (po.name, po.epoch, po.version, po.release, po.arch))
            except yum.Errors.PackageSackError, e:
                # If the package isn't in, search the provides
                pkgs = self.cfg.yumobj.whatProvides(pkg, None, None).returnPackages()
                if len(pkgs) > 0:
                    if not pkg in packages_to_skip:
                        pkgs = self.cfg.yumobj.bestPackagesFromList(pkgs)
                    for po in pkgs:
                        self.cfg.yumobj.tsInfo.addInstall(po)
                        self.log.debug(_("Adding req. pkg %s-%s:%s-%s.%s") % (po.name, po.epoch, po.version, po.release, po.arch))
                else:
                    self.log.error(_("%s. This is a required package.") % e.value, recoverable = False)
                    return False

        return True

    def pkglist_required_installation(self):
        """Add all packages that are required for Installation Media"""
        self.log.debug(_("Adding required packages for Installation Media"))

        if not self.pkglist_required():
            return False

        packages_to_add = self.cfg.required_pkgs_installation + self.cfg.suggested_pkgs_installation
        packages_to_skip = []
        for pkg_tuple in self.cfg.yumobj.tsInfo.pkgdict.keys():
            # Make sure we don't delete any user selected options with "best" package magic
            for package in packages_to_add:
                if str(pkg_tuple[0]) == str(package):
                    self.log.debug(_("Overriding auto package selection with user package selection for %s..." % package))
                    packages_to_skip.append(package)

        for pkg in self.cfg.required_pkgs_installation:
            try:
                pkgs = self.cfg.yumobj.pkgSack.returnNewestByName(pkg)
                for po in pkgs:
                    if not po.name in packages_to_skip:
                        self.cfg.yumobj.tsInfo.addInstall(po)
                        self.log.debug(_("Install Media: Adding required pkg %s-%s:%s-%s.%s") % (po.name, po.epoch, po.version, po.release, po.arch))
            except yum.Errors.PackageSackError, e:
                # If the package isn't in, search the provides
                pkgs = self.cfg.yumobj.whatProvides(pkg, None, None).returnPackages()
                if len(pkgs) > 0:
                    if not pkg in packages_to_skip:
                        pkgs = self.cfg.yumobj.bestPackagesFromList(pkgs)
                    for po in pkgs:
                        self.cfg.yumobj.tsInfo.addInstall(po)
                        self.log.debug(_("Adding req. pkg %s-%s:%s-%s.%s") % (po.name, po.epoch, po.version, po.release, po.arch))
                else:
                    self.log.error(_("%s. This is a required package.") % e.value, recoverable = False)
                    return False

        for pkg in self.cfg.suggested_pkgs_installation:
            try:
                pkgs = self.cfg.yumobj.pkgSack.returnNewestByName(pkg)
                for po in pkgs:
                    self.cfg.yumobj.tsInfo.addInstall(po)
                    self.log.debug(_("Install Media: Adding suggested pkg %s-%s:%s-%s.%s") % (po.name, po.epoch, po.version, po.release, po.arch))
            except yum.Errors.PackageSackError, e:
                pass

        return True

    def pkglist_required_live(self):
        self.log.debug(_("Adding required packages for Live Media"))

        if not self.pkglist_required():
            return False

        packages_to_add = self.cfg.required_pkgs_live + self.cfg.suggested_pkgs_live
        packages_to_skip = []
        for pkg_tuple in self.cfg.yumobj.tsInfo.pkgdict.keys():
            # Make sure we don't delete any user selected options with "best" package magic
            for package in packages_to_add:
                if str(pkg_tuple[0]) == str(package):
                    self.log.debug(_("Overriding auto package selection with user package selection for %s..." % package))
                    packages_to_skip.append(package)

        for pkg in self.cfg.required_pkgs_live:
            try:
                pkgs = self.cfg.yumobj.pkgSack.returnNewestByName(pkg)
#                if len(pkgs) > 1:
#                    pkgs = self.cfg.yumobj.bestPackagesFromList(pkgs)
                for po in pkgs:
                    if not po.name in packages_to_skip:
                        self.cfg.yumobj.tsInfo.addInstall(po)
                        self.log.debug(_("Live Media: Adding req. pkg %s-%s:%s-%s.%s") % (po.name, po.epoch, po.version, po.release, po.arch))
            except yum.Errors.PackageSackError, e:
                # If the package isn't in, search the provides
                pkgs = self.cfg.yumobj.whatProvides(pkg, None, None).returnPackages()
                if len(pkgs) > 0:
                    if not pkg in packages_to_skip:
                        pkgs = self.cfg.yumobj.bestPackagesFromList(pkgs)
                    for po in pkgs:
                        self.cfg.yumobj.tsInfo.addInstall(po)
                        self.log.debug(_("Adding req. pkg %s-%s:%s-%s.%s") % (po.name, po.epoch, po.version, po.release, po.arch))
                else:
                    self.log.error(_("%s. This is a required package.") % e.value)
                    return False

        for pkg in self.cfg.suggested_pkgs_live:
            try:
                pkgs = self.cfg.yumobj.pkgSack.returnNewestByName(pkg)
                for po in pkgs:
                    if not po.name in packages_to_skip:
                        self.cfg.yumobj.tsInfo.addInstall(po)
                        self.log.debug(_("Live Media: Adding suggested pkg %s-%s:%s-%s.%s") % (po.name, po.epoch, po.version, po.release, po.arch))
            except yum.Errors.PackageSackError, e:
                pass

        return True

    def lift_off(self):
        """Do the actual thing. Dance the salsa."""
        self.srpmlist = []
        self.srpmpolist = []
        self.polist = []

        self.cfg.pkglist_selected_tups = self.cfg.yumobj.tsInfo.pkgdict.keys()

        self.log.debug(_("See, this is our list of packages right now: %s") % str(self.cfg.pkglist_selected_tups))

        if self.cfg.media_installation:
            if self.pkglist_required_installation():
                groupList = self.cfg.ksobj._get("packages","groupList")
                packageList = self.cfg.ksobj._get("packages","packageList")
                excludedList = self.cfg.ksobj._get("packages","excludedList")
                self.pkglist_from_ksdata(groupList=groupList, packageList=packageList, excludedList=excludedList, ignore_list=self.cfg.yumobj.tsInfo.pkgdict.keys())
                self.check_dependencies()
                self.download_packages()
                self.buildInstallationMedia()
            else:
                self.log.error(_("Did not succeed in adding in all required packages"), recoverable = False)

#            # If we've done installation media, we want to remove all packages we added
#            # as required packages and that were not in the original list
#            for pkg in self.cfg.yumobj.tsInfo.pkgdict.keys():
#                if not pkg in self.cfg.pkglist_selected_tups:
#                    self.log.debug(_("Removing %s from tsInfo") % str(pkg))
#                    self.cfg.yumobj.tsInfo.remove(pkg)
#                else:
#                    self.log.debug(_("Not removing %s from tsInfo") % str(pkg))

        if self.cfg.media_live:
            self.log.debug(_("Adding Live Media Required Packages"))

            if self.pkglist_required_live():
                groupList = self.cfg.ksobj._get("packages","groupList")
                packageList = self.cfg.ksobj._get("packages","packageList")
                excludedList = self.cfg.ksobj._get("packages","excludedList")
                self.pkglist_from_ksdata(groupList=groupList, packageList=packageList, excludedList=excludedList, ignore_list=[])
                self.check_dependencies(style=self.check_dependencies_no_conflicts)
                self.download_packages()
                self.buildLiveMedia()
            else:
                self.log.error(_("Did not succeed in adding in all required packages"))

        if self.cfg.media_virtual and self.cfg.plugins['modvirt']:
            import revisor.modvirt
            # FIXME: Are we sane?
            # At this point we need to decide if we are going to build the image with yum or kickstart an actual VM using the virt tech
            # Both options should be available
            #self.check_dependencies()
            #self.download_packages()
            self.provision_virtual_machine()

        if self.cfg.revisorUseGUI:
            # Show Finshed Screen
            self.mode.displayFinished()

    def populate_stats(self):
        """ Populate the stats displayed on the ready screen, based on our yum transaction. """
        pbar = self.progress_bar(_("Populating statistics"))
        pkgs = self.cfg.yumobj.tsInfo.getMembers()
        self.cfg.payload_packages = len(pkgs)
        pkgsize = 0
        installsize = 0
        current = 0.0
        total = float(len(pkgs))
        for pkg in pkgs:
            current += 1.0
            try:
                pkgsize += pkg.po.packagesize
            except KeyError, e:
                self.log.debug(_("Package %s-%s:%s-%s.%s does not seem to have a package size header") % (pkg.po.name, pkg.po.epoch, pkg.po.version, pkg.po.release, pkg.po.arch))
            except AttributeError, e:
#                self.log.debug(_("No attribute packagesize for package %s") % pkg.po.name)
                pkgsize += int(pkg.po.returnSimple('packagesize'))

            try:
                installsize += pkg.po.installedsize
            except KeyError, e:
                self.log.debug(_("Package %s-%s:%s-%s.%s does not seem to have a package size header") % (pkg.po.name, pkg.po.epoch, pkg.po.version, pkg.po.release, pkg.po.arch))
            except AttributeError, e:
#                self.log.debug(_("No attribute installedsize for package %s") % pkg.po.name)
                installsize += int(pkg.po.returnSimple('installedsize'))

            pbar.set_fraction(total/current)
        self.cfg.payload_installmedia = pkgsize / 1024 / 1024
        pkgsize = str(self.cfg.payload_installmedia) + " MB"
        self.cfg.payload_livemedia = installsize / 1024 / 1024
        installsize = str(self.cfg.payload_livemedia) + " MB"
        pbar.destroy()

    def buildInstallationMedia(self):
        from revisor import pungi

        # Actually do work.
        if not self.cfg.architecture == 'source':

            # init some
            mypungi = pungi.RevisorPungi(self)

            try:
                os.makedirs(os.path.join(mypungi.destdir,
                                         self.cfg.version,
                                         self.cfg.revisor_model,
                                         self.cfg.architecture))
            except:
                pass

            pkgdir = os.path.join(  mypungi.destdir,
                                    self.cfg.version,
                                    self.cfg.revisor_model,
                                    self.cfg.architecture,
                                    "os",
                                    self.cfg.product_path)

            if not os.path.exists(pkgdir):
                self.log.debug(_("Creating pkgdir: %s" % (pkgdir)))
                os.makedirs(pkgdir)
            else:
                shutil.rmtree(pkgdir, ignore_errors = True)
                os.makedirs(pkgdir)

            if self.cfg.getsource:
                pkgdir_src = os.path.join(  mypungi.destdir,
                                            self.cfg.version,
                                            self.cfg.revisor_model,
                                            "source",
                                            "SRPMS")

                if not os.path.exists(pkgdir_src):
                    self.log.debug(_("Creating pkgdir: %s" % (pkgdir_src)))
                    os.makedirs(pkgdir_src)
                else:
                    shutil.rmtree(pkgdir_src, ignore_errors = True)
                    os.makedirs(pkgdir_src)

            # Now for each txmbr in

            # Link the localPkg() result into the build tree
            pbar = self.progress_bar(_("Linking in packages"))
            failedHardLink = False
            total = float(len(self.polist))
            self.cfg.number_of_packages = total
            i = 0
            pos = self.polist
            if self.cfg.getsource:
                pos += self.srpmpolist
                total += float(len(self.srpmpolist))

            for po in pos:
                if po.arch == "src":
                    link_dir = pkgdir_src
                else:
                    link_dir = pkgdir

                try:
                    os.link(po.localPkg(), link_dir + "/" + os.path.basename(po.localPkg()))
                except OSError, e:
                    self.log.debug(_("Package hard link failed: %s: %s") % (e, link_dir + "/" + os.path.basename(po.localPkg())))
                    if e.errno == 17:
                        continue
                    elif e.errno == 18:
                        if self.cfg.copy_local:
                            self.log.debug(_("Copying: %s to %s") % (po.localPkg(), link_dir + "/" + os.path.basename(po.localPkg())))
                            shutil.copy2(po.localPkg(), link_dir + "/" + os.path.basename(po.localPkg()))
                        else:
                            self.log.debug(_("Symlinking: %s to %s") % (po.localPkg(), link_dir + "/" + os.path.basename(po.localPkg())))
                            try:
                                os.symlink(po.localPkg(), link_dir + "/" + os.path.basename(po.localPkg()))
                                self.log.debug(_("Package symlink succeeded"))
                            except OSError, e:
                                self.log.debug(_("Package link failed, trying copy: %s: %s") % (e, os.path.basename(po.localPkg())))
                                shutil.copy2(po.localPkg(), link_dir + "/" + os.path.basename(po.localPkg()))
                    else:
                        try:
                            self.log.debug(_("Package link failed, trying copy: %s: %s") % (e, os.path.basename(po.localPkg())))
                            shutil.copy2(po.localPkg(), link_dir + "/" + os.path.basename(po.localPkg()))
                        except OSError, e:
                            if e.errno == 17:
                                self.log.debug(_("Package link failed, same file: %s: %s") % (po.localPkg(), link_dir + "/" + os.path.basename(po.localPkg())))
                i += 1
                pbar.set_fraction(i/total)
            pbar.destroy()

            pbar = self.progress_bar(_("Creating Repository Information"), callback = True)
            pungicallback = revisor.progress.PungiCallback(pbar, mypungi)

            database = True
            if os.access("/etc/centos-release", os.R_OK):
                database = False
            elif os.access("/etc/redhat-release", os.R_OK) and not os.access("/etc/fedora-release", os.R_OK):
                database = False

            mypungi.doCreateRepo(database=database,callback=pungicallback)
            pbar.destroy()

            pbar = self.progress_bar(_("Building Installation Images"))
            pungicallback = revisor.progress.PungiCallback(pbar, mypungi)
            mypungi.doBuildinstall(callback=pungicallback)
            pbar.destroy()

            pbar = self.progress_bar(_("Linking in release notes"))
            pungicallback = revisor.progress.PungiCallback(pbar, mypungi)
            mypungi.doGetRelnotes(callback=pungicallback)
            pbar.destroy()

            tree_location = '%s/os' % os.path.join(mypungi.destdir,
                                                                    self.cfg.version,
                                                                    self.cfg.revisor_model,
                                                                    self.cfg.architecture)

            # Add updates.img into the tree
            if self.cfg.updates_img:
                shutil.copy(self.cfg.updates_img, os.path.join(tree_location, 'images/updates.img'))

            if not self.cfg.copy_dir == "":
                if os.access(self.cfg.copy_dir, os.R_OK):
                    mypungi.doCopyDir(copy_dir = self.cfg.copy_dir)
                else:
                    self.log.warning(_("copy_dir '%s' not accessible") % self.cfg.copy_dir)

            calc_size_str = subprocess.Popen(['/usr/bin/du',
                                            '-sbLl',
                                            tree_location],
                                stdout=subprocess.PIPE, stderr=subprocess.STDOUT).communicate()[0]
            calc_size_str = calc_size_str.split()[0]
            self.cfg.mediatypes["unified"]["size"] = int(calc_size_str) + 1
            self.log.debug(_("Size of the installation tree is %s MB") % str(int(calc_size_str) / 1024 / 1024))

            override_retval = ""
            # So, what would the number of discs be?
            for num in self.cfg.mediatypes["index"].keys():
                mediatype = self.cfg.mediatypes["index"][num]
                # Do we even compose this type?
                exec("retval = %s" % self.cfg.mediatypes[mediatype]["compose"])
                if not override_retval == "hihi":
                    self.cfg.mediatypes[mediatype]["compose"] = retval
                else:
                    self.cfg.mediatypes[mediatype]["compose"] = False
                self.cfg.mediatypes[mediatype]["discs"] = int(math.ceil(self.cfg.mediatypes["unified"]["size"] / self.cfg.mediatypes[mediatype]["size"]))
                if self.cfg.mediatypes[mediatype]["discs"] > 1:
                    self.cfg.do_packageorder = True
                else:
                    # Heej if my discnumber is one the rest doesn't have to compose!
                    override_retval = "hihi"

            # Now, we know everything!
            # And we lived happily ever after
            if self.cfg.revisorUseGUI:
                self.mode.BuildMedia.extend_task_list()

# FIXME
# This runs without progress or anything
            if self.cfg.do_packageorder:
                import pkgorder

                # On EL5, the yum cache is placed under the installroot
                # That makes our lives difficult, hihi
                # So, make sure that if we're on el_linux, the repository gets
                # bind mounted and configured appropriately

                pbar = self.progress_bar(_("Running pkgorder"))

                ds = pkgorder.PackageOrderer(self)

# [jeroenm@stevin ~]$ rpm -qv fedora-release; echo $?
# package fedora-release is not installed
# 1
# [jeroenm@stevin ~]$ rpm -qv centos-release; echo $?
# centos-release-5-0.0.el5.centos.2
# 0


                try:
                    os.makedirs(os.path.join(self.cfg.yumobj.cachedir,"anaconda","headers"))
                except:
                    pass

#                el_linux = False
#                if os.access("/etc/centos-release", os.R_OK):
#                    el_linux = True
#                elif os.access("/etc/redhat-release", os.R_OK) and not os.access("/etc/fedora-release", os.R_OK):
#                    el_linux = True
#
#                if el_linux:
#                    src = os.path.join(mypungi.topdir,"os")
#                    dst = ds.testpath + os.path.join(mypungi.topdir)
#                    os.makedirs(dst)
#                    rc = subprocess.call(["/bin/mount", "--bind", src, dest])

                ds.createConfig()

                ds.setup()

                pbar.set_fraction(1.0/28.0)

                ds.printMatchingPkgs("kernel*")
                ds.printMatchingPkgs("iscsi-*")

                pbar.set_fraction(2.0/38.0)

                # Here is where we add groups
                ds.addGroups(["core", "base", "text-internet"])

                pbar.set_fraction(3.0/38.0)
                ds.addGroups(["base-x", "dial-up",
                                "graphical-internet", "editors",
                                "graphics", "gnome-desktop", "sound-and-video",
                                "printing"])

                pbar.set_fraction(11.0/38.0)

                ds.addGroups(["office", "engineering-and-scientific",
                            "authoring-and-publishing", "games"])

                pbar.set_fraction(15.0/38.0)

                ds.addGroups(["web-server", "ftp-server", "sql-server",
                            "mysql", "server-cfg", "dns-server",
                            "smb-server", "admin-tools"])

                pbar.set_fraction(23.0/38.0)

                ds.addGroups(["kde-desktop", "development-tools", "development-libs",
                            "gnome-software-development", "eclipse",
                            "x-software-development",
                            "java-development", "kde-software-development",
                            "mail-server", "network-server", "legacy-network-server"])

                pbar.set_fraction(34.0/38.0)

                ds.addGroups(["news-server", "legacy-software-development"])

                # Everthing else but kernels
                for po in ds.pkgSack.returnPackages():
                    fpattern = "%s*" % po.name
                    ds.printMatchingPkgs(fpattern)

                pbar.destroy()

#                if el_linux:
#                    rc = subprocess.call(["/bin/umount", dst])

            if self.cfg.media_virtual:
                # Make our kickstart auto installing
                # FIXME: For some reason the install still asks for user input during partitioning.
                self.cfg.kickstart_include = True
                self.cfg.kickstart_default = True
                self.cfg.ksobj._set("autopart",val=True)
                self.cfg.ksobj._set("clearpart","initAll",True)
                self.cfg.ksobj._set("interactive",val=False)
                self.cfg.ksobj._set("mediacheck",val=False)
                self.cfg.ksobj._set("reboot","action",kickstart.constants.KS_SHUTDOWN)
                self.cfg.ksobj._set("reboot","eject",True)
                self.cfg.ksobj._set("firstboot",val=True)
                if self.cfg.ksobj._get("rootpw","password") == "":
                    self.cfg.ksobj._set("rootpw","password","revisor")
                    self.cfg.ksobj._set("rootpw","isCrypted",val=False)

            if self.cfg.kickstart_include:
# FIXME
                # This needs some more work because the kickstart file may have been
                # generated during our run, partially modified during our run, and may not have
                # been used at all
                if self.cfg.use_kickstart_file:
                    shutil.copyfile(self.cfg.kickstart_file, os.path.join(mypungi.topdir,"ks.cfg"))
                else:
                    f = open(os.path.join(mypungi.topdir,"ks.cfg"), "w")
                    f.write(self.cfg.ksobj.__str__())
                    f.close()
                if self.cfg.kickstart_default:
                    f = open(os.path.join(mypungi.topdir,"isolinux","isolinux.cfg"),"rw+")
                    buf = f.read()
#                    buf = buf.replace("#prompt 1","prompt 1")
                    buf = buf.replace("menu default\n","")
                    buf = buf.replace("default linux","default ks")
#                    buf = buf.replace("default vesamenu.c32","default ks")
                    f.seek(0)
                    f.write(buf)
                    f.write("label ks\n  menu label " + _("^Install using kickstart") + "\n  menu default\n  kernel vmlinuz\n  append initrd=initrd.img ks=cdrom:/ks.cfg %s" % self.cfg.im_append)
                    f.close()

# Do some really neat routine here
            for num in self.cfg.mediatypes["index"].keys():
                mediatype = self.cfg.mediatypes["index"][num]

                discs = self.cfg.mediatypes[mediatype]["discs"]
                size = self.cfg.mediatypes[mediatype]["size"]
                discdir = self.cfg.mediatypes[mediatype]["discdir"]
                label = self.cfg.mediatypes[mediatype]["label"]

                if self.cfg.mediatypes[mediatype]["compose"]:
                    os.symlink(mypungi.topdir,"%s-%s" % (mypungi.topdir, discdir))

                    for i in range(2, discs+1):
                        try:
                            os.makedirs("%s-%s-disc%d" % (mypungi.topdir, discdir, i))
                        except:
                            pass

                    if discs > 1:
                        # Split Tree
                        pbar = self.progress_bar(_("Splitting Build Tree (%s)") % label)
                        pungicallback = revisor.progress.PungiCallback(pbar, mypungi)
                        mypungi.doSplittree(media_size=size,discdir=discdir,callback=pungicallback)
                        pbar.destroy()

                        # Split repo
                        pbar = self.progress_bar(_("Splitting Repository (%s)") % label)
                        pungicallback = revisor.progress.PungiCallback(pbar, mypungi)
                        mypungi.doCreateSplitrepo(discdir=discdir,callback=pungicallback)
                        pbar.destroy()

                    # Create ISOs
                    if discs > 1:
                        for disc in range(1, discs+1):
                            pbar = self.progress_bar(_("Creating %s ISO Image #%d") % (label, disc))
                            pungicallback = revisor.progress.PungiCallback(pbar, mypungi)
                            mypungi.doCreateIso(mediatype=mediatype, disc=disc, callback=pungicallback)
                            pbar.destroy()
                    else:
                        pbar = self.progress_bar(_("Creating %s ISO Image") % label)
                        pungicallback = revisor.progress.PungiCallback(pbar, mypungi)
                        mypungi.doCreateIso(mediatype=mediatype, callback=pungicallback)
                        pbar.destroy()

            # Do some SHA1SUMMONING
            pbar = self.progress_bar(_("Creating SHA1SUMs for Images"))
            pbar.set_fraction(0.0)
            total = float(len(self.cfg.built_iso_images))
            cur = 0.0
            for built_image in self.cfg.built_iso_images:
                isofile = os.path.basename(built_image["location"])
                pbar.set_markup(isofile)
                isodir = os.path.dirname(built_image["location"])
                # shove the sha1sum into a file
                sha1file = open(os.path.join(isodir,"SHA1SUM"), 'a')
                self.run_command(['/usr/bin/sha1sum', isofile], rundir=isodir, output=sha1file)
                sha1file.close()
                cur += 1.0
                pbar.set_fraction(cur/total)
            pbar.destroy()

# Prep for delta image creation
            if self.cfg.delta_old_image and self.cfg.plugins['moddelta']:
                from revisor.moddelta import RevisorDelta
                delta = RevisorDelta(self)
                # Generate the delta image
                # Find the types of images if self.cfg.delta_old_image is a directory
                delta.build_iso_deltas()

# Prep for cobbler
            if self.cfg.plugins['modcobbler']:
                if self.cfg.cobbler_add_profile:
                    pbar = self.progress_bar(_("Adding Profile to Cobbler"))
                    # FIXME: This should be a simple, quick call; do we need a callback?
                    start = float(1.0)
                    end = float(10.0)
                    pbar.set_fraction(start/end)
                    self.log.debug(_(("Using Cobbler to add profile %s") % (self.cfg.cobbler_add_profile)))
                    kickstart = os.path.join(self.cfg.working_directory,"%s.cfg" % (self.cfg.cobbler_add_profile))
                    f = open(kickstart, "w")
                    f.write(self.cfg.ksobj.__str__())
                    f.close()
                    self.cfg.cobbler.newProfile(self.cfg.cobbler_add_profile, self.cfg.cobbler_use_distro, kickstart)
                    pbar.set_fraction(start)
                    pbar.destroy()

                if self.cfg.cobbler_add:
                    pbar = self.progress_bar(_("Adding Distro and Profile to Cobbler"))
                    cobblercallbackhack = revisor.progress.CobblerCallbackHack(pbar, self.cfg)
                    pbar.set_fraction(float(0.5))
                    self.log.debug(_(("Using Cobbler to add distro and profile %s") % (self.cfg.cobbler_add)))
                    self.cfg.cobbler.addCompose(str(mypungi.topdir), str(self.cfg.cobbler_add), callback=cobblercallbackhack)
                    pbar.set_fraction(float(0.9))
                    pbar.destroy()

        if self.cfg.getsource:
            tree_dst = os.path.join(self.cfg.destination_directory,"os","source")
            tree_src = os.path.join(mypungi.destdir,self.cfg.version,self.cfg.revisor_model,"source")
            self.log.debug(_("Moving %s to %s") % (tree_src,tree_dst))
            shutil.move(tree_src,tree_dst)

        if self.cfg.media_installation_tree:
            tree_dst = os.path.join(self.cfg.destination_directory,"os",self.cfg.architecture)
            tree_src = mypungi.topdir
            self.log.debug(_("Moving %s to %s") % (tree_src,tree_dst))
            shutil.move(tree_src,tree_dst)

        shutil.rmtree(os.path.join(self.cfg.working_directory,"revisor-pungi"))

    def buildLiveMedia(self):
        livemedia_targetsize = self.cfg.payload_livemedia + 1536
        self.log.debug(_("Setting Live Media Ext3 Filesystem Size to %s MB") % str(livemedia_targetsize))
        # Get our object from pilgrim
        liveImage = lm_optical.LMOptical(self)

        # Symlink our yum.installroot to
        liveImage.fs_label = "livecd-" + time.strftime("%Y%m%d")
        liveImage.bindmounts = []
        liveImage.tmpdir = self.cfg.working_directory

        pbar = self.progress_bar(_("Setting up Ext3 Filesystem"))

        liveImage.setup(livemedia_targetsize, build_dir="/var/tmp/revisor-livecd", yum_installroot=self.cfg.yumobj.conf.installroot, yum_cachedir=self.cfg.yumobj.conf.cachedir)
        self.log.info(_("Successfully set up the installation target for live media."))

        pbar.destroy()

        pbar = self.progress_bar(_("Installing Packages"), callback = True)
        liveImage.installPackages(pbar = pbar)

        self.cfg.yumobj.closeRpmDB()

        if not self.cfg.copy_dir == "":
            if os.access(self.cfg.copy_dir, os.R_OK):
                # So what is the destination directory?
                src_dir = self.cfg.copy_dir
                dst_dir = "%s/install_root/%s"

                for root, dirs, files in os.walk(src_dir):
                    real_dst = dst_dir % (liveImage.build_dir, root.replace(src_dir,''))
                    if not os.access(real_dst, os.R_OK):
                        self.log.debug(_("Creating %s") % real_dst)
                        os.makedirs(real_dst)

                    for file in files:
                        self.log.debug(_("Copying %s to %s") % (os.path.join(root,file), os.path.join(real_dst,file)))
                        shutil.copyfile(os.path.join(root,file), os.path.join(real_dst,file))

            else:
                self.log.warning(_("copy_dir '%s' not accessible") % self.cfg.copy_dir)

        pbar.destroy()

        pbar = self.progress_bar(_("Configuring Network"))
        liveImage.configureNetwork()
        pbar.destroy()

        pbar = self.progress_bar(_("Configure System"))
        liveImage.configureSystem()
        liveImage.extra_configuration()
        liveImage.runPost()
        pbar.destroy()

        pbar = self.progress_bar(_("Creating RAM Filesystem"))
        liveImage.createInitramfs()
        pbar.destroy()

        pbar = self.progress_bar(_("Relabel System"))
        liveImage.relabelSystem()
        pbar.destroy()

        pbar = self.progress_bar(_("Configure BootLoader"))
        liveImage.configureBootloader()
        pbar.destroy()

        if self.cfg.lm_chroot_shell:
            self.log.info(_("Launching an interactive shell in the live image. 'exit' when you are finished:"))
            # FIXME: Find a better way? Fix liveImage.launchShell() ?
            os.system("/usr/sbin/chroot %s/install_root" % (liveImage.build_dir))

        if self.cfg.advanced_configuration:
            pbar = self.progress_bar(_("Launching Shell"))
            if self.cfg.revisorUseGUI:
                pbar.set_markup(_("Launching a Shell. Adjust the configuration from within the Chroot and logout to continue building the Live Media."))
                subprocess.call(["/usr/bin/gnome-terminal", "--window", "-x", "chroot", "%s/install_root" % (liveImage.build_dir)])
            else:
                liveImage.launchShell() # Doesn't seem to work?
            pbar.destroy()

        pbar = self.progress_bar(_("Unmounting Filesystems"))
        liveImage.unmount(kill = False)
        pbar.destroy()

        if not self.cfg.lm_ignore_deleted:
            liveImage.cleanupDeleted()

        if not self.cfg.lm_skip_fs_compression:
            pbar = self.progress_bar(_("Creating Squash Filesystem"))
            pungipbar = revisor.progress.PungiCallback(pbar)
            pbar.set_markup(_("This is going to take a while and might cause the GUI to become unresponsive... it will recover after this task completes."))
            liveImage.createSquashFS(pungipbar)
            pbar.destroy()

        pbar = self.progress_bar(_("Creating ISO Live Image"))
        pungipbar = revisor.progress.PungiCallback(pbar)
        liveImage.createIso(pungipbar)
        pbar.destroy()

        pbar = self.progress_bar(_("Cleaning up build environment"))
        liveImage.teardown(kill = False)
        pbar.destroy()

    def set_run_command_callback(self, callback=None):
        self.run_command_callback = callback

    def run_command(self, command, rundir=None, output=subprocess.PIPE, error=subprocess.STDOUT, inshell=False, env=None, callback=None):
        """Run a command and log the output."""

        if rundir == None:
            rundir = os.path.join(self.cfg.working_directory,"revisor-rundir")
        self.log.debug(_("Setting rundir to %s") % rundir)

        if not os.access(rundir, os.R_OK):
            try:
                os.makedirs(rundir)
            except:
                self.log.debug(_("Directory %s could not be created. Aborting"), recoverable = False)

        self.log.debug(_("Running command: %s") % ' '.join(command))
        self.log.debug(_("Extra information: %s %s %s") % (rundir, inshell, env))

        if output == subprocess.PIPE:
            p1 = subprocess.Popen(command, cwd=rundir, stdout=output, stderr=error, shell=inshell, env=env)

            while p1.poll() == None:
                try:
                    returncode = int(p1.poll())
                except:
                    if output == subprocess.PIPE:
                        try:
                            line = p1.stdout.readline()
                            if not line == "":
                                callback.parse_line(command, line)
                        except:
                            pass
        else:
            p1 = subprocess.Popen(command, cwd=rundir, stdout=output, stderr=error)
            (out, err) = p1.communicate()

    def provision_virtual_drive(self, partlayout=None):
        """ Setup a virtual drive to setup a virtual machine inside. Returns the drive object."""

        fslabel = "revisor"
        virtual_type = ""

        if not self.cfg.virt_xen == "":
            virtual_type = "xen"
            lofile = os.path.join(self.cfg.destination_directory, 'virt-xen', '%s.img' % self.cfg.virt_xen)
            mountdir = os.path.join(self.cfg.working_directory, 'virt-xen-%s' % self.cfg.virt_xen)
            fslabel = self.cfg.virt_xen
            size = self.cfg.virt_xen_size
        elif not self.cfg.virt_kvm == "":
            virtual_type = "kvm"
            lofile = os.path.join(self.cfg.destination_directory, 'virt-kvm', '%s.img' % self.cfg.virt_kvm)
            mountdir = os.path.join(self.cfg.working_directory, 'virt-kvm-%s' % self.cfg.virt_kvm)
            fslabel = self.cfg.virt_kvm
            size = self.cfg.virt_kvm_size
        else:
            print "Oh noes!"
            sys.exit(1)

        virt_gear_drive = revisor.modvirt.RevisorLoopbackMount(virtual_type, lofile, mountdir, size, fslabel, self.cfg.virt_fstype, self.cfg)
        virt_gear_drive.partlayout = partlayout
        return virt_gear_drive

    def provision_virtual_machine(self, virtual_drive=None):
        """ Setup and build a virtual machine. """
        size = 0
        if not self.cfg.virt_xen == "":
            size = self.cfg.virt_xen_size
        elif not self.cfg.virt_kvm == "":
            size = self.cfg.virt_kvm_size

        if self.cfg.virt_appliance:
            # Use a raw disk and yum, with some other magic, to build the guest image.

            #FIXME: Actually build the part data dict
            partdata = None
            #partdata = self.cfg.ksobj._set("something","somethingelse",value)

            #FIXME: If size is not 0, make sure our partition layout will fit in the requested size or bail

            if not virtual_drive:
                virtual_drive = self.provision_virtual_drive(partdata)
                virtual_drive.setup()

            virtual_drive.mount_drive(offset=32256)
            #virtual_drive.mount_drive()
            virtual_drive.prime_drive()
            install_root = virtual_drive.mountdir
            print install_root
        else:
            # Use the virtual technology to run an auto installing ISO image and build the virtual guest.
            virtual_type = ""
            name = ""
            install_memory = 386
            vcpus = 1
            if not self.cfg.virt_xen == "":
                virtual_type = "xen"
                name = self.cfg.virt_xen
            elif not self.cfg.virt_kvm == "":
                virtual_type = "kvm"
                name = self.cfg.virt_kvm

            # Get a virtual drive to install onto
            virtual_drive = self.provision_virtual_drive()
            virtual_disk = virtual_drive.get_disk()

            # Get the ISO image to install from
            iso_image = ""
            for image in self.cfg.built_iso_images:
                if image["type"] == "dvd": # Maybe we need to create a special ISO image for this task?
                    iso_image = image["location"]

            virtual_machine = None
            #FIXME: Setup a better server connection then just the default
            server = revisor.modvirt.virtinst.util.default_connection()
            conn = revisor.modvirt.libvirt.open(server)

            # FIXME: We should really test some of the following:
            # modvirt.virtinst.util. [...]
            # is_pae_capable()
            # is_hvm_capable()
            # is_kqemu_capable()
            # is_kvm_capable()
            # is_blktap_capable()

            if virtual_type is "xen":
                # Setup the virtual machine
                # FIXME: (#235) try:
                virtual_machine = revisor.modvirt.RevisorParaVirtual(type=virtual_type, connection = conn)
                # except:
            elif virtual_type is "kvm":
                # Setup the virtual machine
                # FIXME: (#235) try:
                virtual_machine = revisor.modvirt.RevisorFullVirtual(type=virtual_type, connection = conn)
                # except:

            if not virtual_machine:
                self.log.error(_("Unable to create virtual machine instance."), recoverable = False)

            virtual_machine.set_log(self.log)
            virtual_machine.disks.append(virtual_disk)
            virtual_machine.iso = iso_image
            virtual_machine.set_name(name)
            virtual_machine.set_memory(install_memory)
            virtual_machine.set_vcpus(vcpus)

            ## FIXME: We might want to connect the user to the install session if we are in GUI mode? Offer info about vnc if in CLI mode?
            # Progress bars are going to be a PITA for this.
            virtual_machine.set_graphics("vnc") # So we can connect while testing and see what is going wrong/right
            ##

            virtual_machine.start_virtual_machine_installer()

