#
# Copyright 2007 Fedora Unity
#
# Jonathan Steffan <jon a fedoraunity.org>
# Jeroen van Meeuwen <kanarip a fedoraunity.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 only
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import sys
import os
from revisor.constants import *
from revisor.errors import *
import shutil
import subprocess

from ConfigParser import SafeConfigParser

import yum
from yum.constants import *
import yum.Errors
import logging
from rhpl.translate import _, N_, getDefaultLangs
import rhpl.translate as translate

import kickstart
import pykickstart
try:
    import pykickstart.errors
    import pykickstart.version
except:
    pass

class ConfigStore:
    """ This ConfigStore holds all options available in Revisor, including Kickstart Data and Handlers,
    so that these are available to both the GUI and the CLI mode."""
    def __init__(self, cli_options, log):
        self.log = log
        self.cli_options = cli_options
        self.revisorUseGUI = cli_options.revisorUseGUI
        self.revisorUseCLI = cli_options.revisorUseCLI
        self.check_plugins()

    def check_plugins(self):
        test_plugins = ['modcobbler','moddelta','modvirt','modjigdo','modrebrand']
        self.plugins = {}

        for plugin in test_plugins:
            try:
                self.plugins[plugin] = True
                exec("import revisor.%s" % plugin)
            except:
                self.plugins[plugin] = False

        self.log.debug(str(self.plugins))

    def setup_cfg(self):
        self.options_set_defaults()

        if self.cli_options.revisorUseCLI or self.cli_options.revisorUseGUI:
            self.options_set_from_config(self.cli_options)
            self.options_set_from_commandline(self.cli_options)

        # These packages get added anyway
        self.required_pkgs =                [
                                                'grub',
                                                'kernel',
                                                'mailx',
                                                'nano',
                                                'selinux-policy',
                                                'selinux-policy-targeted',
                                                'yum'
                                            ]

        # These packages get added to installation media (required)
        self.required_pkgs_installation =   [
                                                'acl',
                                                'anaconda',
                                                'anaconda-runtime',
                                                'attr',
                                                'authconfig',
                                                'busybox-anaconda',
                                                'dejavu-lgc-fonts',
                                                'dosfstools',
                                                'firstboot',
                                                'fontconfig',
                                                'glibc',
                                                'gnome-python2-gtkhtml2',
                                                'hdparm',
                                                'iscsi-initiator-utils',
                                                'joe',
                                                'kernel-xen',
                                                'man',
                                                'mdadm',
                                                'memtest86+',
                                                'openssl',
                                                'pango',
                                                'policycoreutils',
                                                'syslinux',
                                                'system-config-keyboard',
                                                'xorg-x11-drivers',
                                                'xorg-x11-fonts-ISO8859-1-75dpi'
                                            ]

        # These packages get added to installation media (if available)
        self.suggested_pkgs_installation =  [
                                                'jfsutils',
                                                'reiserfs-utils',
                                                'xfsprogs',
                                                'xorg-x11-fonts-base'
                                            ]


        # These packages get added to live media (required)
        self.required_pkgs_live =           [
                                                'authconfig',
                                                'cpuspeed',
                                                'eject',
                                                'file',
                                                'man',
                                                'nano',
                                                'openssh-clients',
                                                'passwd',
                                                'rootfiles',
                                                'rpm',
                                                'rsync',
                                                'sendmail',
                                                'shadow-utils',
                                                'sudo',
                                                'syslinux',
                                                'tree',
                                                'wget',
                                                'xterm'
                                            ]

        # These packages get added to live media (if available)
        self.suggested_pkgs_live =          [
                                                'system-logos'
                                            ]

        # Then there's also packages that are not required, or suggested for that matter,
        # but that need to not match a certain arch, and all archs should be added for.
        # One example is libbeagle (i386, i486, i586, i686)
        self.archs_pkgs_installation =      [
                                                'libbeagle',
                                                'pm-utils'
                                            ]

        # Nasty, but it saves us from requiring system-config-display/firstboot
        self.default_xorg_config = """

Section "ServerLayout"
        Identifier     "Default Layout"
        Screen      0  "Screen0" 0 0
        InputDevice    "Keyboard0" "CoreKeyboard"
EndSection

Section "InputDevice"
        Identifier  "Keyboard0"
        Driver      "kbd"
        Option      "XkbModel" "pc105"
        Option      "XkbLayout" "%s"
EndSection

Section "Device"
        Identifier  "Videocard0"
        Driver      "%s"
EndSection

Section "Screen"
        Identifier "Screen0"
        Device     "Videocard0"
        DefaultDepth     24
        SubSection "Display"
                Viewport   0 0
                Depth     24
        EndSubSection
EndSection

"""


    def check_package_selection(self, only_clear_data=False):
        """ A function to check some key things about our selected packages. """

        self.package_Xorg = False
        self.package_windowmanager_gnome = False
        self.package_windowmanager_kde = False
        self.package_windowmanager_xfce = False

        if not only_clear_data:
            pkgs = self.yumobj.tsInfo.getMembers()
            for pk in pkgs:
                if pk.name in ('xorg-x11-server-Xorg',):
                    self.package_Xorg = True
                elif pk.name in ('kdebase',):
                    self.package_windowmanager_kde = True
                elif pk.name in ('gnome-session',):
                    self.package_windowmanager_gnome = True
                elif pk.name in ('xfce-session',):
                    self.package_windowmanager_xfce = True

##
##
## Functions that can load and test regular configuration files
##
##

    def set_revisor_config(self,val):
        self.revisor_config = val

    def set_revisor_model(self,val):
        self.revisor_model = val

    def set_main_conf(self,val):
        self.main_conf = val

    def check_working_directory(self):
        """From the Revisor Configuration Files and other settings,
        check for the directory hierarchy"""

        self.log.debug(_("Checking working directories"))
        complain = False
        if os.access(os.path.join(self.working_directory,"pungi-revisor"), os.R_OK):
            complain = True
        if os.access(os.path.join(self.working_directory,"revisor-pungi"), os.R_OK):
            complain = True
        if os.access(os.path.join(self.working_directory,"revisor"), os.R_OK):
            complain = True
        if os.access(os.path.join(self.working_directory,"revisor-rundir"), os.R_OK):
            complain = True
        if os.access(os.path.join(self.working_directory,"revisor-livecd"), os.R_OK):
            complain = True

        if complain:
            if not self.answer_yes:
                self.log.warning(_("The directories Revisor uses in %s already exist. This could possibly hold data from a previous run. Please remove or move them to a safe location, then confirm to continue. If you do not move or remove the files, Revisor will simply delete them." % self.working_directory))
            else:
                self.log.debug(_("The directories Revisor uses in %s already exist. Revisor deleted them." % self.working_directory))
            complain = False
            if os.access(os.path.join(self.working_directory,"pungi-revisor"), os.R_OK):
                shutil.rmtree(os.path.join(self.working_directory,"pungi-revisor"))
            if os.access(os.path.join(self.working_directory,"revisor-pungi"), os.R_OK):
                shutil.rmtree(os.path.join(self.working_directory,"revisor-pungi"))
            if os.access(os.path.join(self.working_directory,"revisor"), os.R_OK):
                if os.path.ismount("%s%s" % (self.working_directory,"revisor")):
                    for f in ["/sys", "/proc", "/dev/pts", "/dev", "/selinux", "/yum-cache", ""]:
                        if os.path.ismount("%s%s%s" % (self.working_directory,"revisor",f)):
                            self.log.debug(_("%s%s%s is a mount, trying to unmount") % (self.working_directory,"revisor",f))
                            try:
                                rc = subprocess.call(["/bin/umount", "%s%s%s" % (self.working_directory,"revisor",f)])
                            except OSError, e:
                                self.log.error(_("Unable to unmount %s%s%s with error:\n\n%s\n\nPlease resolve the issue and continue.") % (self.working_directory,"revisor",f,e))
                                return False
                        else:
                            self.log.debug(_("%s%s%s is not a mount") % (self.working_directory,"revisor",f))

                    buf = subprocess.Popen(["/sbin/losetup", "-a"], stdout=subprocess.PIPE).communicate()[0]
                    for line in buf.split("\n"):
                        # loopdev: fdinfo (filename)
                        fields = line.split()
                        if len(fields) != 3:
                            continue
                        if "data/os.img" in fields[2] and "revisor" in fields[2]:
                            loopdev = fields[0][:-1]
                            rc = subprocess.call(["/sbin/losetup", "-d", loopdev])
                            break

                try:
                    shutil.rmtree("%s%s" % (self.working_directory,"revisor"))
                    # Then, recreate them
                    os.makedirs("%s/%s" % (self.working_directory,"revisor/etc"))
                    # And copy in our main_conf
                    shutil.copyfile(self.main_conf,"%s/%s" % (self.working_directory,"revisor/etc/yum.conf"))
                except OSError, e:
                    self.log.error(_("Unable to unmount %s%s with error:\n\n%s\n\nPlease resolve the issue and continue.") % (self.working_directory,"revisor",e))
                    return False

            if os.access(os.path.join(self.working_directory,"revisor-rundir"), os.R_OK):
                shutil.rmtree(os.path.join(self.working_directory,"revisor-rundir"))
            if os.access(os.path.join(self.working_directory,"revisor-livecd"), os.R_OK):
                shutil.rmtree(os.path.join(self.working_directory,"revisor-livecd"))

        return True

    def check_destination_directory(self):
        """From the Revisor Configuration Files and other settings,
        check for the directory hierarchy"""

        self.log.debug(_("Checking destination directories"))

        self.destination_directory = os.path.join(self.destination_directory,self.revisor_model)

        self.log.debug(_("Set destination directory to %s") % self.destination_directory)

        if os.access(self.destination_directory, os.R_OK):
            complain = False
            if os.access(os.path.join(self.destination_directory,"iso"), os.R_OK):
                if int(self.media_installation_cd) + int(self.media_installation_dvd) + int(self.media_installation_dvd_duallayer) + int(self.media_installation_unified) > 0:
                    complain = True
            if os.access(os.path.join(self.destination_directory,"live"), os.R_OK):
                if self.media_live:
                    complain = True
            if os.access(os.path.join(self.destination_directory,"xen"), os.R_OK):
                if self.media_virtual_xen:
                    complain = True
            if os.access(os.path.join(self.destination_directory,"kvm"), os.R_OK):
                if self.media_virtual_kvm:
                    complain = True
            if os.access(os.path.join(self.destination_directory,"os"), os.R_OK):
                if self.media_installation_tree or self.getsource:
                    complain = True

            if complain:
                if not self.answer_yes:
                    self.log.warning(_("The directories Revisor uses in %s already exist. This could possibly hold data from a previous run. Please remove or move them to a safe location, then confirm to continue. If you do not move or remove the files, Revisor will simply delete them." % self.destination_directory))
                else:
                    self.log.debug(_("The directories Revisor uses in %s already exist. Revisor deleted them." % self.destination_directory))
                complain = False
                if os.access(os.path.join(self.destination_directory,"iso"), os.R_OK):
                    if int(self.media_installation_cd) + int(self.media_installation_dvd) + int(self.media_installation_dvd_duallayer) + int(self.media_installation_unified) > 0:
                        shutil.rmtree(os.path.join(self.destination_directory,"iso"))
                if os.access(os.path.join(self.destination_directory,"live"), os.R_OK):
                    if self.media_live:
                        shutil.rmtree(os.path.join(self.destination_directory,"live"))
                if os.access(os.path.join(self.destination_directory,"xen"), os.R_OK):
                    if self.media_virtual_xen:
                        shutil.rmtree(os.path.join(self.destination_directory,"xen"))
                if os.access(os.path.join(self.destination_directory,"kvm"), os.R_OK):
                    if self.media_virtual_kvm:
                        shutil.rmtree(os.path.join(self.destination_directory,"kvm"))
                if os.access(os.path.join(self.destination_directory,"os"), os.R_OK):
                    if self.media_installation_tree or self.getsource:
                        shutil.rmtree(os.path.join(self.destination_directory,"os"))
#        else:
#            os.makedirs(self.destination_directory)

        if not os.access(os.path.join(self.destination_directory,"iso"), os.R_OK) and int(self.media_installation_cd) + int(self.media_installation_dvd) + int(self.media_installation_dvd_duallayer) + int(self.media_installation_unified) > 0:
            while not os.access(os.path.join(self.destination_directory,"iso"), os.R_OK):
                try:
                    os.makedirs(os.path.join(self.destination_directory,"iso"))
                except:
                    self.log.error(_("Cannot access %s, please check the permissions so we can try again." % os.path.join(self.destination_directory,"iso")))

        if not os.access(os.path.join(self.destination_directory,"live"), os.R_OK) and self.media_live:
            while not os.access(os.path.join(self.destination_directory,"live"), os.R_OK):
                try:
                    os.makedirs(os.path.join(self.destination_directory,"live"))
                except:
                    self.log.error(_("Cannot access %s, please check the permissions so we can try again." % os.path.join(self.destination_directory,"live")))

        if not os.access(os.path.join(self.destination_directory,"xen"), os.R_OK) and self.media_virtual_xen:
            while not os.access(os.path.join(self.destination_directory,"xen"), os.R_OK):
                try:
                    os.makedirs(os.path.join(self.destination_directory,"xen"))
                except:
                    self.log.error(_("Cannot access %s, please check the permissions so we can try again." % os.path.join(self.destination_directory,"xen")))

        if not os.access(os.path.join(self.destination_directory,"kvm"), os.R_OK) and self.media_virtual_kvm:
            while not os.access(os.path.join(self.destination_directory,"kvm"), os.R_OK):
                try:
                    os.makedirs(os.path.join(self.destination_directory,"kvm"))
                except:
                    self.log.error(_("Cannot access %s, please check the permissions so we can try again." % os.path.join(self.destination_directory,"kvm")))

        if not os.access(os.path.join(self.destination_directory,"os"), os.R_OK) and ( self.media_installation_tree or self.getsource ):
            while not os.access(os.path.join(self.destination_directory,"os"), os.R_OK):
                try:
                    os.makedirs(os.path.join(self.destination_directory,"os"))
                except:
                    self.log.error(_("Cannot access %s, please check the permissions so we can try again." % os.path.join(self.destination_directory,"os")))

##
##
##  Yum related stuff
##
##

    def setup_yum(self):
        # You do like having your own package sack, don't you?
        self.yumobj = yum.YumBase()

        if self.architecture == "i386":
            self.arch_list = yum.rpmUtils.arch.getArchList('i686')
            self.yumobj.compatarch = 'i686'
        elif self.architecture == "ppc":
            self.arch_list = yum.rpmUtils.arc.getArchList('ppc64')
            self.yumobj.compatarch = 'ppc64'
        else:
            self.arch_list = yum.rpmUtils.arch.getArchList(self.architecture)
            self.yumobj.compatarch = self.architecture

    def reposSetup(self, callback = None, thisrepo = None):
        # We shouldn't run for all keys(), but for all repos
        tmpparser = SafeConfigParser()
        tmpparser.read(self.main_conf)

        if not self.repo_override:
            for item in tmpparser._sections:
                # Maybe if item in self.repos.keys()... but that fuck up matching development against
                # extras-development... Right?
                try:
                    r = self.repos[item]
#                    # Override source repositories if getsource is True
#                    if item[:-6] == "source" and self.getsource:
#                        r = True
                    if r:
                        self.yumobj.repos.enableRepo(item)
                    else:
                        self.yumobj.repos.disableRepo(item)
                except:
                    pass

        if callback:
            self.yumobj.repos.callback = callback
            callback.num_tasks += 10

        self.yumobj.doLock(YUM_PID_FILE)

        if callback: callback.next_task()
        try:
            if hasattr(self.yumobj,"_getRepos"):
                self.yumobj._getRepos(thisrepo)
            else:
                self.log.debug(_("Using deprecated YUM function: %s()") % "doRepoSetup")
                self.yumobj.doRepoSetup(thisrepo)
        except yum.Errors.RepoError, e:
            raise RevisorDownloadError, e

        if callback: callback.next_task()

        try:
            if hasattr(self.yumobj,"_getGroups"):
                self.yumobj._getGroups()
            else:
                self.log.debug(_("Using deprecated YUM function: %s()") % "doGroupSetup")
                self.yumobj.doGroupSetup()
        except yum.Errors.GroupsError, e:
            self.log.warning(_("No groups present! Error was: %s") % e)
        except yum.Errors.RepoError, e:
            raise RevisorDownloadError, e

        if callback: callback.next_task(next = 5) # hack... next should be long

        try:
            if hasattr(self.yumobj,"_getSacks"):
                self.yumobj._getSacks(archlist=self.arch_list)
            else:
                self.log.debug(_("Using deprecated YUM function: %s()") % "doSackSetup")
                self.yumobj.doSackSetup(archlist=self.arch_list)
        except yum.Errors.RepoError, e:
            raise RevisorDownloadError, e

        if callback: callback.next_task(incr=3)

        if callback: self.yumobj.repos.callback = None

##
##
## Kickstart related stuff
##
##

    def test_ks(self, file):
        ksobj_test = kickstart.RevisorKickstart(self)
        ksobj_test.create_parser()

        try:
            ksobj_test.read_file(file)
            return True
        except:
            return False

    def setup_ks(self):
        self.ksobj = kickstart.RevisorKickstart(self)
        self.ksobj.create_parser()

    def load_kickstart(self, fn, packages=True):
        """Function loads a kickstart from a file, yes it does.
        Set the 'packages' parameter to include or exclude the
        package manifest"""
        self.kickstart_file = fn
        self.ksobj.read_file(fn)

    def set_kickstart_file(self,val):
        self.kickstart_file = val

    def config_to_kickstart(self,ks):
        """This would be a function that writes the configstore
        out to a pykickstart object"""
        pass

##
##
## Revisor Configuration related stuff
##
##

    def options_set_defaults(self):
        """Sets the default configuration options for Revisor"""

        self.log.debug(_("Setting default options"))

        # Default, ask user what they want to do
        self.answer_yes = False

        # Media Type defaults for Advanced Configuration
        self.media_installation = False
        self.media_live = False
        self.media_virtualization = False
        self.media_utility = False

        # Defaults for Installation Media
        self.media_installation_cd = False
        self.media_installation_dvd = False
        self.media_installation_dvd_duallayer = False
        self.media_installation_tree = False
        self.media_installation_raw = False
        self.media_installation_unified = False
        self.media_installation_pxe = False
        self.updates_img = ""

        # Defaults for Live Media
        self.media_live_optical = False
        self.media_live_thumb = False
        self.media_live_hd = False
        self.media_live_raw = False

        # Defaults for Virtualization Media
        self.media_virtual = False
        self.media_virtual_vmware_appliance = False
        self.media_virtual_vmware_guest = False
        self.media_virtual_xen = False
        self.media_virtual_kvm = False

        # Defaults for Utility Media
        self.media_utility_rescue = False

        # Default configuration file. Customize via BASE_CONF_DIR in
        # revisor.constants.py
        self.revisor_config = os.path.join(BASE_CONF_DIR,"revisor.conf")

        # Default destination directory
        self.destination_directory = "/srv/revisor/"

        # Default working directory
        self.working_directory = "/var/tmp/"

        # Sets which models we want to use for spinning
        # Either have an empty string here (takes 1 line) or have a lot
        # of useless code doing hasattr() and try/except (You'll see this often)
        self.main_conf = ""
        self.revisor_model = "f7-i386"

        # Show repos, hide repos
        self.repos_enablesource = False
        self.repos_enabledebuginfo = False
        self.repos_enabletesting = False
        self.repos_enabledevelopment = False

        # Options! More options!
        self.kickstart_file = os.path.join(BASE_CONFD_DIR,"sample-ks.cfg")
        self.use_kickstart_file = True

        # Use the package manifest from the kickstart file you provided earlier
        self.kickstart_manifest = False

        # Customize that package manifest. Works like a charm if packages have
        # become obsolete or fail dependencies.
        self.kickstart_manifest_customize = False
        self.kickstart_options_customize = False
        self.kickstart_include = False
        self.kickstart_default = False
        self.kickstart_repos = True

        # Show advanced configuration. Allows for manual package version
        # selection, kickstart options override, chroot shell, etc.
        self.advanced_configuration = False

        # Sets the mode in which we depsolve:
        # Allow conflicts between the packages for install media, or
        # don't allow them so you can install everything?
        self.dependency_resolve_allow_conflicts = False

        # So what mode are we in? Are we doing customization, or are we doing respins?
        # This has impact on the way we use ksdata to addInstall() packages (actually, matches of ksdata)
        self.mode_respin = False

        # Should Revisor copy files instead of trying to link them, and finally symlink them? Note that
        # this has a major impact on disk space, especially when using local file:// repositories
        # However, this also helps you to maintain a source tree for distributing your product
        self.copy_local = False

        # Default values overridden by config of model options
        self.product_name = 'Fedora'
        self.product_path = 'Fedora'
        self.iso_label = 'Fedora'
        self.iso_basename = 'F'
        self.comps = '/etc/revisor/comps-f7.xml'
        # Needs to be a string
        self.version = "7"
        self.getsource = 0
        self.copy_dir = ''
        self.strip_copy_dir = ''

        self.rebrand = False
        self.rebrand_packages = "fedora-release fedora-release-notes fedora-logos"
        self.rebrand_directory = "/usr/share/revisor/rebrand-empty"
        self.release = "1"
        self.release_url = "http://fedoraproject.org"
        self.release_codename = "Moonshine"

        # Media sizes, in bytes
        self.mediatypes =   {   "index":
                                            {
                                                0: "cd",
                                                1: "dvd",
                                                2: "dvd_dl",
                                                3: "unified"
                                            },
                                "cd":       {
                                                "size": 685 * 1024 * 1024,
                                                "discdir": "cd",
                                                "label": "CD",
                                                "discs": 0,
                                                "compose": "self.cfg.media_installation_cd"
                                            },
                                "dvd":      {
                                                "size": 4300 * 1024 * 1024,
                                                "discdir": "dvd",
                                                "label": "DVD",
                                                "discs": 0,
                                                "compose": "self.cfg.media_installation_dvd"
                                            },
                                "dvd_dl":   {
                                                "size": 2 * 4300 * 1024 * 1024,
                                                "discdir": "dvd-dl",
                                                "label": "DVD-DL",
                                                "discs": 0,
                                                "compose": "self.cfg.media_installation_dvd_duallayer"
                                            },
                                "unified":  {
                                                "size": -1,
                                                "discdir": "unified",
                                                "label" : "",
                                                "discs": 1,
                                                "compose": "self.cfg.media_installation_unified"
                                            }
                            }

        # Use fedora-release by default
        self.release_pkgs = "^fedora-release.*$"
        self.release_files = "eula.txt fedora.css GPL README-BURNING-ISOS-en_US.txt RELEASE-NOTES-en_US.html ^RPM-GPG img images stylesheet-images"

        # Installation Media Options
        # Append this when kickstart is set to boot by default:
        self.im_append = ''

        # Live Media options
        self.lm_use_shadow = True
        self.lm_use_md5 = True
        self.lm_use_nis = False
        self.lm_use_ldap = False
        self.lm_use_kerberos = False
        self.lm_use_hesiod = False
        self.lm_use_samba = False
        self.lm_use_nscd = False
        self.lm_chroot_shell = False

        self.lm_user_configuration = False
        self.lm_user_name = "live"
        self.lm_user_comment = "Fedora Unity - Revisor"
        self.lm_user_password = "live"

        self.lm_user_auto_login = False
        self.lm_user_wheel = True
        self.lm_wheel_sudo_no_passwd = True
        self.lm_dump_current_profile = False

        self.lm_preferred_kernel_type = "normal"

        self.lm_fs_label = "fc"
        self.lm_skip_fs_compression = False
        self.lm_skip_prelink = False
        self.lm_ignore_deleted = True

        # short, long, options
        self.lm_bootloader_stanzas = [("linux",      _("Run from image"),                "rhgb quiet"),
                                      ("runfromram", _("Run from RAM - requires 1 GB+"), "rhgb quiet live_ram")]

        self.lm_bootloader_timeout = 600
        self.lm_bootloader_isolinux_extras = ""
        self.lm_hostname = "livecd.fedoraunity.org"
        self.lm_default_runlevel = 3

        # Repo store for dialog population
        self.repos = {}

        # Default architecture we're running for
        self.architecture = "i386"

        # Still reading?
        self.i_did_all_this = False
        self.nis_auth = ""
        self.ldap_auth = ""
        self.kerberos_auth = ""
        self.hesiod_auth = ""
        self.samba_auth = ""
        self.nscd_auth = ""

        self.tasks = []
        self.built_iso_images = []
        self.repo_override = []

        # Cobbler Actions
        self.cobbler_add_profile = ""
        self.cobbler_add_distro = ""
        self.cobbler_use_distro = ""
        self.cobbler_use_profile = ""
        self.cobbler_from_distro = ""
        self.cobbler_add = ""
        self.cobbler_list = False
        self.cobbler_server = ""

        # Delta Options
        self.delta_old_image = ""

        # Virtual Image Options
        self.virt_xen = ""
        self.virt_xen_size = 3000
        self.virt_kvm = ""
        self.virt_kvm_size = 3000
        self.virt_sparse = False
        self.virt_stateless = False
        self.virt_appliance = False
        self.virt_fstype = "ext3"

        # Init package booleans
        self.check_package_selection(only_clear_data=True)

        # Track how many packages we are dealing with, useful for progressbars
        self.number_of_packages = 0

        # Set to not do package order by default
        self.do_packageorder = False

        # Set a lot of bullshit
        self.cmd_mkisofs = ['/usr/bin/mkisofs', '-v', '-U', '-J', '-R', '-T', '-f']

        self.x86bootargs = ['-b', 'isolinux/isolinux.bin', '-c', 'isolinux/boot.cat',
                            '-no-emul-boot', '-boot-load-size', '4', '-boot-info-table']

        self.ia64bootargs = ['-b', 'images/boot.img', '-no-emul-boot']

        self.ppcbootargs = ['-part', '-hfs', '-r', '-l', '-sysid', 'PPC', '-no-desktop', '-allow-multidot', '-chrp-boot',
                            '-map', os.path.join('/usr/lib/anaconda-runtime/boot', 'mapping'),
                            '-magic', os.path.join('/usr/lib/anaconda-runtime/boot', 'magic'),
                            '-hfs-bless']

    def options_reload_from_config(self):
        config = SafeConfigParser()
        config.read(self.revisor_config)

        if config.has_section("revisor"):
            for (key,val) in config.items("revisor"):
                self.log.debug("setting %s to %s" % (key,val))
                exec("self.%s = %s" % (key,val))

    def model_options_reload_from_config(self):
        config = SafeConfigParser()
        config.read(self.revisor_config)

        if config.has_option(self.revisor_model,"main"):
            self.main_conf = config.get(self.revisor_model,"main")

        for varname in self.__dict__.keys():
#            self.log.debug(_("Testing for configuration parameter %s") % varname)
            if not config.has_option(self.revisor_model,varname):
                continue

            # These are illegal variables
            if varname in ["revisor_model", "revisor_config", "i_did_all_this"]:
                continue

            if isinstance(self.__dict__[varname], int):
                val = config.getint(self.revisor_model,varname)
            elif isinstance(self.__dict__[varname], bool):
                val = config.getboolean(self.revisor_model,varname)
            elif isinstance(self.__dict__[varname], str):
                val = config.get(self.revisor_model,varname)
            elif isinstance(self.__dict__[varname], list):
                val = eval(config.get(self.revisor_model,varname))

            if hasattr(self,"check_setting_%s" % varname):
                exec("retval = self.check_setting_%s(%r)" % (varname, val))
                if not retval:
                    continue

            setattr(self,varname,val)
            self.log.debug(_("Setting %s to %r (from configuration file model %s)") % (varname,val,self.revisor_model))

    def check_setting_main(self, val):
        if val == "/etc/yum.conf":
            self.log.error(_("You cannot choose the system's yum configuration file /etc/yum.conf for use with Revisor. Aborting."), recoverable = False)

    def check_setting_iso_label(self, val):
        if len(val) > 32:
            self.log.warning(_("The ISO label cannot be longer then 32 characters due to Joliet limitations"))
            return False
        else:
            return True

    def check_setting_kickstart_file(self, val):
        if not os.access(val, os.R_OK):
            self.log.error(_("Kickstart file %s cannot be read.") % val)
            return False
        else:
            self.test_ks(val)
            return True

    def check_setting_comps(self, val):
        if not os.access(val, os.R_OK):
            self.log.warning(_("The file configured as a comps file (%s) cannot be read. When composing installation media, this is FATAL.") % val)
            return False
        else:
            return True

    def options_set_from_config(self, cli_options):
        """Sets the default configuration options for Revisor from a
        configuration file. Configuration file may be customized using
        the --config CLI option"""

        self.log.debug(_("Setting options from configuration file"))

        # Check from which configuration file we should get the defaults
        if cli_options.revisor_config:
            self.log.debug(_("Using configuration file %s, as specified by --config") % cli_options.revisor_config)
            self.revisor_config = cli_options.revisor_config

        config = SafeConfigParser()
        self.log.debug(_("Reading configuration file %s") % self.revisor_config)
        config.read(self.revisor_config)

        if config.has_section("revisor"):
            if config.has_option("revisor","model"):
                if config.has_section(config.get("revisor","model")):
                    self.revisor_model = config.get("revisor","model")

            for varname in self.__dict__.keys():
                if not config.has_option("revisor",varname):
                    continue

                # These are illegal variables
                if varname in ["revisor_model", "revisor_config", "i_did_all_this"]:
                    continue

                if isinstance(self.__dict__[varname], int):
                    val = config.getint("revisor",varname)
                elif isinstance(self.__dict__[varname], bool):
                    val = config.getboolean("revisor",varname)
                elif isinstance(self.__dict__[varname], str):
                    val = config.get("revisor",varname)

                if hasattr(self,"check_setting_%s" % varname):
                    exec("retval = self.check_setting_%s(%r)" % (varname, val))
                    if not retval:
                        continue

                setattr(self,varname,val)
                self.log.debug(_("Setting %s to %r (from configuration file)") % (varname,val))

        if cli_options.revisor_model:
            self.log.debug(_("Using model %s, as specified by --model") % cli_options.revisor_model)
            self.revisor_model = cli_options.revisor_model

        if not self.revisor_model == None and not self.revisor_model == "":
            if not config.has_section(self.revisor_model):
                if self.revisorUseCLI:
                    self.log.error(_("No such model %s in configuration file %s") % (self.revisor_model,self.revisor_config), recoverable = False)
                else:
                    self.log.info(_("No such model %s in configuration file %s") % (self.revisor_model,self.revisor_config))
            elif config.has_option(self.revisor_model,"main"):
                if os.access(config.get(self.revisor_model,"main"), os.R_OK):
                    self.main_conf = config.get(self.revisor_model,"main")
                else:
                    self.log.error(_("Invalid model %s, configuration file %s not readable" % (self.revisor_model,config.get(self.revisor_model,"main"))), recoverable = self.revisorUseGUI)
            else:
                self.log.error(_("Invalid model %s, no configuration file %s" % (self.revisor_model,"main")), recoverable = self.revisorUseGUI)

            if config.has_option(self.revisor_model,"pungi"):
                # If this directive is here... We're using old configuration files
                self.log.error(_("You may be using an old configuration file, since there is a 'pungi' configuration directive in %s" % self.revisor_model))
            else:
                # Read the configuration directives from the general revisor configuration file's model section
                self.model_options_reload_from_config()

    def options_set_from_commandline(self, cli_options):
        """Overrides default options from the CLI"""

        self.log.debug(_("Setting options from command-line"))
        self.revisorUseCLI = cli_options.revisorUseCLI
        self.revisorUseGUI = cli_options.revisorUseGUI

        config = SafeConfigParser()
        config.read(self.revisor_config)

        if self.revisorUseCLI:
            if self.revisor_model == "" or self.revisor_model == None:
                # No default, model should be specified on CLI, or be loaded from config
                if not cli_options.revisor_model == None:
                    self.revisor_model = cli_options.revisor_model
                else:
                    self.log.error(_("No model specified."), recoverable = False)
            else:
                # I don't know what else. Tests for the model come after this little routine.
                pass

        if self.revisorUseCLI:
            if not config.has_section(self.revisor_model):
                self.log.error(_("Invalid model specified: %s does not exist in %s") % (self.revisor_model,self.revisor_config), recoverable = False)

        # Add options that need to be saved after setting defaults here:
        if not cli_options.kickstart_file == self.kickstart_file and not cli_options.kickstart_file == "":
            self.kickstart_file = cli_options.kickstart_file
        if cli_options.kickstart_include:
            self.kickstart_include = cli_options.kickstart_include
        if cli_options.kickstart_default:
            self.kickstart_default = cli_options.kickstart_default

        if not cli_options.destination_directory == self.destination_directory:
            self.destination_directory = cli_options.destination_directory
        if not cli_options.working_directory == self.working_directory:
            self.working_directory = cli_options.working_directory

        if cli_options.media_installation_dvd:
            self.media_installation_dvd = cli_options.media_installation_dvd
        if cli_options.media_installation_dvd_duallayer:
            self.media_installation_dvd_duallayer = cli_options.media_installation_dvd_duallayer
        if cli_options.media_installation_cd:
            self.media_installation_cd = cli_options.media_installation_cd
        if cli_options.media_installation_tree:
            self.media_installation_tree = cli_options.media_installation_tree
        if cli_options.media_installation_raw:
            self.media_installation_raw = cli_options.media_installation_raw
        if cli_options.media_installation_unified:
            self.media_installation_unified = cli_options.media_installation_unified
        if not cli_options.updates_img == "":
            self.updates_img = cli_options.updates_img

        if cli_options.media_live_optical:
            self.media_live_optical = cli_options.media_live_optical
        if cli_options.media_live_thumb:
            self.media_live_thumb = cli_options.media_live_thumb
        if cli_options.media_live_hd:
            self.media_live_hd = cli_options.media_live_hd
        if cli_options.media_live_raw:
            self.media_live_raw = cli_options.media_live_raw
        if cli_options.lm_chroot_shell:
            self.lm_chroot_shell = cli_options.lm_chroot_shell
        if cli_options.lm_skip_fs_compression:
            self.lm_skip_fs_compression = cli_options.lm_skip_fs_compression
        if cli_options.lm_skip_prelink:
            self.lm_skip_prelink = cli_options.lm_skip_prelink
        if cli_options.lm_ignore_deleted:
            self.lm_ignore_deleted = cli_options.lm_ignore_deleted
        if cli_options.lm_preferred_kernel:
            if cli_options.lm_preferred_kernel in ["normal", "PAE", "xen", "debug"]:
                self.lm_preffered_kernel = cli_options.lm_preferred_kernel
            else:
                self.log.error(_("Preferred kernel should be one of: normal, PAE, xen, debug."), recoverable = False)


        if not cli_options.cobbler_add_profile == "":
            self.cobbler_add_profile = cli_options.cobbler_add_profile
        if not cli_options.cobbler_use_distro == "":
            self.cobbler_use_distro = cli_options.cobbler_use_distro
        if not cli_options.cobbler_use_profile == "":
            self.cobbler_use_profile = cli_options.cobbler_use_profile
#            if not cli_options.cobbler_from_distro == "":
#                self.cobbler_from_distro = cli_options.cobbler_from_distro
        if not cli_options.cobbler_add == "":
            self.cobbler_add = cli_options.cobbler_add
        if cli_options.cobbler_list:
            self.cobbler_list = cli_options.cobbler_list
        if not cli_options.cobbler_server == "":
            self.cobbler_server = cli_options.cobbler_server

        if not cli_options.delta_old_image == "":
            self.delta_old_image = cli_options.delta_old_image

        if cli_options.answer_yes:
            self.answer_yes = cli_options.answer_yes

        if not cli_options.virt_xen == "":
            self.virt_xen = cli_options.virt_xen
        if not cli_options.virt_xen_size == "3000":
            self.virt_xen_size = cli_options.virt_xen_size

        if not cli_options.virt_kvm == "":
            self.virt_kvm = cli_options.virt_kvm
        if not cli_options.virt_kvm_size == "3000":
            self.virt_kvm_size = cli_options.virt_kvm_size

        if cli_options.virt_sparse:
            self.virt_sparse = cli_options.virt_sparse
        if cli_options.virt_stateless:
            self.virt_stateless = cli_options.virt_stateless
        if not cli_options.virt_fstype == "ext3":
            self.virt_fstype = cli_options.virt_fstype
        if cli_options.virt_appliance:
            self.virt_appliance = cli_options.virt_appliance

        if cli_options.mode_respin:
            self.mode_respin = cli_options.mode_respin

        if cli_options.getsource:
            self.getsource = cli_options.getsource

        if self.media_installation_dvd or self.media_installation_dvd_duallayer or self.media_installation_cd or self.media_installation_tree or self.media_installation_unified or self.media_installation_raw:
            self.media_installation = True
        else:
            self.media_installation = False

        if self.media_live_thumb or self.media_live_optical or self.media_live_hd or self.media_live_raw:
            self.media_live = True
        else:
            self.media_live = False

        if not self.virt_xen == "":
            self.media_virtual = True
            self.media_virtual_xen = True

        if not self.virt_kvm == "":
            self.media_virtual = True
            self.media_virtual_kvm = True

        if not self.virt_kvm == "" and not self.virt_xen == "":
            self.log.error(_("Currently, building Xen and KVM guests at the same time is not supported. Soon. Really."), recoverable = False)

        if self.media_installation_raw or self.media_live_raw:
            self.log.error(_("Currently, building build raw images is not supported. Soon. Really."), recoverable = False)

        if self.media_virtual and self.media_installation:
            self.log.warning(_("You have selected to both build an installer image and a virtual machine. A side effect is that your ISO image will automagically run the installation."))

        if not self.virt_appliance and self.media_virtual:
            self.media_installation = True
            self.media_installation_unified = True

        # FIXME
        # Disabling Virtualization Media for release
        if self.media_virtual:
            self.log.error(_("Virtualization Media is not yet available, we're sorry for the inconvenience"), recoverable = False)

        # FIXME
        # Disabling raw image generation for release
        if self.media_installation_raw or self.media_live_raw:
            self.log.error(_("Raw images are not yet available, we're sorry for the inconvenience"), recoverable = False)

        if self.updates_img and not self.media_installation:
            self.log.error(_("Updates.img is only usable with installation media. Please remove --updates-img or also build installation media."), recoverable = False)
        elif self.updates_img and not os.access(self.updates_img, os.R_OK):
            self.log.error(_("The updates.img %s is not accessible." % self.updates_img), recoverable = False)

        if self.revisorUseCLI and not self.cobbler_use_profile and not self.kickstart_file:
            self.log.error(_("No kickstart file specified"), recoverable = False)

        if self.kickstart_file:
            if self.cobbler_use_profile:
                self.use_kickstart_file = True
            elif os.access(self.kickstart_file, os.R_OK):
                self.use_kickstart_file = True
                self.kickstart_file = self.kickstart_file
            elif self.revisorUseCLI:
                self.log.error(_("Kickstart file %s fails to load" % self.kickstart_file), recoverable = False)
            else:
                self.log.error(_("Kickstart file %s fails to load, continuing with defaults" % self.kickstart_file))

        if self.revisorUseCLI and not self.media_installation and not self.media_live and not self.media_virtual:
            self.log.error(_("No media specified"), recoverable = False)

        if self.mode_respin and int(self.media_installation) + int(self.media_live) + int(self.media_virtual) > 1:
            self.log.error(_("When composing in respin mode, you can only select one type of media install, live or virtual. You can however select multiple media such as cd, dvd, etc."), recoverable = False)

#        if self.media_installation and self.media_live:
#            self.log.error(_("We're sorry to inform you that Revisor at this moment is unable to spin both Installation Media and Live Media in the same run. Please refer to http://revisor.fedoraunity.org/development/install-and-live-media/ for more information."), recoverable = False)

        if not self.plugins['modcobbler']:
            # FIXME: there has to be a better way to do the below
            if self.cobbler_add_profile or self.cobbler_add or self.cobbler_use_distro or self.cobbler_list or self.cobbler_use_profile or self.cobbler_server:
                self.log.error(_("Revisor Cobbler plugin options specified, but no cobbler module was found. Install revisor-cobbler."), recoverable = False)
        else:
            # FIXME: Maybe we need to test if we need cobbler or not. For now, it loads if we have it.
            import revisor.modcobbler
            self.cobbler = revisor.modcobbler.RevisorCobbler()
            if self.cobbler_server:
                if not self.cobbler_use_distro and not self.cobbler_use_profile:
                    self.log.error(_("Specifying a Cobbler server is only supported for read-only actions. Currently, --cobbler-use-distro and --cobbler-use-profile."), recoverable = False)
            if self.cobbler_add_profile and not (self.cobbler_use_distro or self.cobbler_from_distro):
                self.log.error(_("Add profile was selected without selecting what distro to use. Add --cobbler-use-distro or --cobbler-from-distro."), recoverable = False)
            if self.cobbler_use_distro:
                distro_found = False
                if not self.cobbler_server:
                    distros = self.cobbler.distros()
                    for distro in distros:
                        if distro.name == self.cobbler_use_distro:
                            self.repo_override.append({"name": distro.name, "baseurl": distro.ks_meta["tree"], "gpgcheck": 0})
                            distro_found = True
                else:
                    #FIXME: Add try: except: with a nice error message
                    port = 25151
                    server_url = "http://%s:%s" % (self.cobbler_server, port)
                    self.cobblerserver = revisor.modcobbler.RevisorCobblerServer(server_url)
                    distros = self.cobblerserver.get_distros()
                    for distro in distros:
                        if distro["name"] == self.cobbler_use_distro:
                            distro_found = True

                if not distro_found:
                    self.log.error(_("The distro '%s' does not exist." % (self.cobbler_use_distro)), recoverable = False)

            if self.cobbler_use_profile:
                kickstart_url = ""

                if not self.cobbler_server:
                    profiles = self.cobbler.profiles()
                    profile_obj = False
                    for profile in profiles:
                        if profile.name == self.cobbler_use_profile:
                            profile_obj = profile

                    if not profile_obj:
                        self.log.error(_("The profile '%s' does not exist." % (self.cobbler_use_profile)), recoverable = False)

                    kickstart = os.path.join(profile_obj.settings.webdir, 'kickstarts', profile_obj.name, 'ks.cfg')
                    if not os.access(kickstart, os.R_OK):
                        kickstart = 'http://%s/cblr/kickstarts/%s/ks.cfg' % (profile_obj.server, profile_obj.name)

                    kickstart_url = kickstart

                else:
                    #FIXME: Add try: except: with a nice error message
                    port = 25151
                    server_url = "http://%s:%s" % (self.cobbler_server, port)
                    self.cobblerserver = revisor.modcobbler.RevisorCobblerServer(server_url)
                    profiles = self.cobblerserver.get_profiles()
                    profile_obj = False
                    for profile in profiles:
                        if profile["name"] == self.cobbler_use_profile:
                            profile_obj = profile

                    if not profile_obj:
                        self.log.error(_("The profile '%s' does not exist." % (self.cobbler_use_profile)), recoverable = False)

                    kickstart = 'http://%s/cblr/kickstarts/%s/ks.cfg' % (self.cobbler_server, profile_obj["name"])

                    kickstart_url = kickstart

                self.kickstart_file = kickstart_url

        if not self.plugins['moddelta']:
            if self.delta_old_image:
                self.log.error(_("Delta image creation specified, but no delta module found. Install revisor-delta."), recoverable = False)

        elif self.delta_old_image:
            if not os.access(self.delta_old_image, os.R_OK):
                self.log.error(_(("%s could not be found, please specify a valid ISO.") % (options.delta_old_image)), recoverable = False)
            if not self.media_installation_cd and not self.media_installation_dvd and not self.media_installation_unified:
                self.log.error(_("To be able to create a delta, you need to generate a new ISO image. Add --install-cd, --install-dvd or --install-unified."), recoverable = False)
            if os.path.isdir(self.delta_old_image):
                self.log.debug(_("'delta_old_image' is a directory, so we can delta both CD sets and DVD sets"))
            else:
                self.log.debug(_("'delta_old-image' is not a directory, so we can only delta a CD or a DVD, not both or more then one"))
                if int(self.media_installation_cd) + int(self.media_installation_dvd) + int(self.media_installation_dvd_duallayer) + int(self.media_installation_unified) > 1:
                    self.log.error(_("More then one image specified, while --delta-iso-image doesn't point to a directory."), recoverable = False)

        if not self.plugins['modvirt']:
            if self.media_virtual:
                self.log.error(_("Virtualization options specified, but no virt module found. Install revisor-virt."), recoverable = False)

