#
# livecd-creator : Creates Live CD based for Fedora.
#
# Copyright 2007, Red Hat  Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import getopt
import os
import os.path
import glob
import sys
import errno
import re
import string
import tempfile
import time
import traceback
import subprocess
import shutil
import yum
import rpmUtils.arch

import revisor.kickstart as kickstart

# Translation
from rhpl.translate import _, N_, getDefaultLangs
import rhpl.translate as translate

class MountError(Exception):
    def __init__(self, msg):
        Exception.__init__(self, msg)

class InstallationError(Exception):
    def __init__(self, msg):
        Exception.__init__(self, msg)

class BindChrootMount:
    """Represents a bind mount of a directory into a chroot."""
    def __init__(self, src, chroot, dest = None):
        self.src = src
        self.root = chroot

        if not dest:
            dest = src
        self.dest = self.root + "/" + dest

        self.mounted = False

    def mount(self):
        if not self.mounted:
            if not os.path.exists(self.dest):
                os.makedirs(self.dest)
            rc = subprocess.call(["/bin/mount", "--bind", self.src, self.dest])
            if rc != 0:
                raise MountError(_("Bind-mounting '%s' to '%s' failed") % (self.src, self.dest))
            self.mounted = True

    def umount(self):
        if self.mounted:
            rc = subprocess.call(["/bin/umount", self.dest])
            self.mounted = False


class LoopbackMount:
    def __init__(self, lofile, mountdir, fstype = None):
        self.lofile = lofile
        self.mountdir = mountdir
        self.fstype = fstype

        self.mounted = False
        self.losetup = False
        self.rmdir   = False
        self.loopdev = None

    def cleanup(self):
        self.umount()
        self.lounsetup()

    def umount(self):
        if self.mounted:
            rc = subprocess.call(["/bin/umount", self.mountdir])
            self.mounted = False

        if self.rmdir:
            try:
                os.rmdir(self.mountdir)
            except OSError, e:
                pass
            self.rmdir = False

    def lounsetup(self):
        if self.losetup:
            rc = subprocess.call(["/sbin/losetup", "-d", self.loopdev])
            self.losetup = False
            self.loopdev = None

    def loopsetup(self):
        if self.losetup:
            return

        rc = subprocess.call(["/sbin/losetup", "-f", self.lofile])
        if rc != 0:
            raise MountError(_("Failed to allocate loop device for '%s'") % self.lofile)

        # succeeded; figure out which loopdevice we're using
        buf = subprocess.Popen(["/sbin/losetup", "-a"],
                               stdout=subprocess.PIPE).communicate()[0]
        for line in buf.split("\n"):
            # loopdev: fdinfo (filename)
            fields = line.split()
            if len(fields) != 3:
                continue
            if fields[2] == "(%s)" %(self.lofile,):
                self.loopdev = fields[0][:-1]
                break

        if not self.loopdev:
            raise MountError(_("Failed to find loop device associated with '%s' from '/sbin/losetup -a'") % self.lofile)

        self.losetup = True

    def mount(self):
        if self.mounted:
            return

        self.loopsetup()

        if not os.path.isdir(self.mountdir):
            os.makedirs(self.mountdir)
            self.rmdir = True

        args = [ "/bin/mount", self.loopdev, self.mountdir ]
        if self.fstype:
            args.extend(["-t", self.fstype])

        rc = subprocess.call(args)
        if rc != 0:
            raise MountError(_("Failed to mount '%s' to '%s'") % (self.loopdev, self.mountdir))

        self.mounted = True

class SparseExt3LoopbackMount(LoopbackMount):
    def __init__(self, lofile, mountdir, size, fslabel, tmpdir):
        LoopbackMount.__init__(self, lofile, mountdir, fstype = "ext3")
        self.size = size
        self.fslabel = fslabel

    def _createSparseFile(self):
        dir = os.path.dirname(self.lofile)
        if not os.path.isdir(dir):
            os.makedirs(dir)

        # create the sparse file
        fd = os.open(self.lofile, os.O_WRONLY | os.O_CREAT)
        off = long(self.size * 1024L * 1024L)
        os.lseek(fd, off, 0)
        os.write(fd, '\x00')
        os.close(fd)

    def _formatFilesystem(self):
        rc = subprocess.call(["/sbin/mkfs.ext3", "-F", "-L", self.fslabel,
                              "-m", "1", "-b", "4096", self.lofile,
                              "%d" % ( self.size *  1024L * 1024L / 4096L )])
        if rc != 0:
            raise MountError(_("Error creating ext3 filesystem"))
        rc = subprocess.call(["/sbin/tune2fs", "-c0", "-i0", "-Odir_index",
                              "-ouser_xattr,acl", self.lofile])

    def mount(self):
        self._createSparseFile()
        self._formatFilesystem()
        return LoopbackMount.mount(self)

class InstallationTarget:
    def write_fstab(self):
        fstab = open(self.build_dir + "/install_root/etc/fstab", "w")
        fstab.write("/dev/mapper/livecd-rw   /                       ext3    defaults,noatime 0 0\n")
        fstab.write("devpts                  /dev/pts                devpts  gid=5,mode=620  0 0\n")
        fstab.write("tmpfs                   /dev/shm                tmpfs   defaults        0 0\n")
        fstab.write("proc                    /proc                   proc    defaults        0 0\n")
        fstab.write("sysfs                   /sys                    sysfs   defaults        0 0\n")
        fstab.close()

    def setup(self, image_size, base_on = None, build_dir = None, yum_installroot = None, yum_cachedir=None):
        """setup target ext3 file system in preparation for an install"""

        # setup temporary build dirs
        if not build_dir:
            try:
                self.build_dir = tempfile.mkdtemp(dir="/var/tmp", prefix="livecd-creator-")
            except OSError, (err, msg):
                raise InstallationError(_("Failed create build directory in /var/tmp: %s") % msg)
        else:
            self.build_dir = build_dir
            os.makedirs(self.build_dir)

        os.makedirs(self.build_dir + "/out/isolinux")
        os.makedirs(self.build_dir + "/out/sysroot")
        os.makedirs(self.build_dir + "/data/sysroot")

        # Replace making the install_root directory with linking it
        # to the existing install_root, if specified
        if not yum_installroot:
            os.makedirs(self.build_dir + "/install_root")
        else:
            os.symlink(yum_installroot,self.build_dir + "/install_root")

        # Create a yum cache directory only if needed, or set it to the correct directory
        if not yum_cachedir:
            os.makedirs(self.build_dir + "/yum-cache")
# Don't know if this is usefull for others, revisor doesn't need it
#        else:
#            os.symlink(yum_cachedir, self.build_dir + "/yum-cache")

        if base_on:
            # get backing ext3 image if we're based this build on an existing live CD ISO
            self.base_on_iso(base_on)

            self.instloop = LoopbackMount("%s/data/os.img" %(self.build_dir,),
                                          "%s/install_root" %(self.build_dir,))
        else:
            self.instloop = SparseExt3LoopbackMount("%s/data/os.img"
                                                    %(self.build_dir,),
                                                    "%s/install_root"
                                                    %(self.build_dir,),
                                                    image_size,
                                                    self.fs_label,
                                                    self.tmpdir)


        try:
            self.instloop.mount()
        except MountError, e:
            raise InstallationError(_("Failed to loopback mount '%s' : %s") % (self.instloop.lofile, e))

        if not base_on:
            # create a few directories needed if it's a new image
            os.makedirs(self.build_dir + "/install_root/etc")
            os.makedirs(self.build_dir + "/install_root/boot")
            os.makedirs(self.build_dir + "/install_root/var/log")
            os.makedirs(self.build_dir + "/install_root/var/cache/yum")

        # bind mount system directories into install_root/
        for (f, dest) in [("/sys", None), ("/proc", None), ("/dev", None),
                          ("/dev/pts", None), ("/selinux", None)]:
            self.bindmounts.append(BindChrootMount(f, self.build_dir + "/install_root", dest))

        if not yum_cachedir:
            self.bindmounts.append(BindChrootMount(self.build_dir + "/yum-cache", self.build_dir + "/install_root", "/var/cache/yum"))
            cachedir = "/var/cache/yum"
        else:
            cachedir = yum_cachedir

        for b in self.bindmounts:
            b.mount()

        # make sure /etc/mtab is current inside install_root
        os.symlink("../proc/mounts", self.build_dir + "/install_root/etc/mtab")

        self.write_fstab()

        return True

    def unmount(self, kill = True):
        """detaches system bind mounts and install_root for the file system and tears down loop devices used"""

        # FIXME: Make one last ditch effort to close fds still open
        # in the install root; this is only needed because
        # there's no way to ask yum to close its sqlite dbs,
        # though. See https://bugzilla.redhat.com/236409
        if kill:
            for i in range(3, os.sysconf("SC_OPEN_MAX")):
                try:
                    os.close(i)
                except:
                    pass

        try:
            os.unlink(self.build_dir + "/install_root/etc/mtab")
        except OSError:
            pass

        self.bindmounts.reverse()
        for b in self.bindmounts:
            b.umount()

        if self.instloop:
            (xxbsize, xxfrsize, xxblocks, xxbfree,
             xxbavail, xxfiles, xxffree, xxfavail,
             xxflag, xxnamemax) = os.statvfs(self.build_dir + "/install_root")
            print >> sys.stderr, _("Installation target uncompressed data size is %d MB") % ( ( xxfrsize * xxblocks - xxbsize * xxbfree ) / ( 1024L * 1024L ) )
            self.instloop.cleanup()
            self.instloop = None

    def teardown(self, kill = True):
        if self.build_dir:
            self.unmount(kill = kill)
            shutil.rmtree(self.build_dir, ignore_errors = True)

    def run_in_root(self):
        print "%s/install_root" %(self.build_dir,)
        os.chroot("%s/install_root" %(self.build_dir,))
        os.chdir("/")

    def writeNetworkIfCfg(self, instroot, network):
        path = instroot + "/etc/sysconfig/network-scripts/ifcfg-" + network.device

        f = file(path, "w+")
        os.chmod(path, 0644)

        f.write("DEVICE=%s\n" % network.device)
        f.write("BOOTPROTO=%s\n" % network.bootProto)

        if network.bootProto.lower() == "static":
            if network.ip:
                f.write("IPADDR=%s\n" % network.ip)
            if network.netmask:
                f.write("NETMASK=%s\n" % network.netmask)

        if network.onboot:
            f.write("ONBOOT=on\n")
        else:
            f.write("ONBOOT=off\n")

        if network.essid:
            f.write("ESSID=%s\n" % network.essid)

        if network.ethtool:
            if network.ethtool.find("autoneg") == -1:
                network.ethtool = "autoneg off " + network.ethtool
            f.write("ETHTOOL_OPTS=%s\n" % network.ethtool)

        if network.bootProto.lower() == "dhcp":
            if network.hostname:
                f.write("DHCP_HOSTNAME=%s\n" % network.hostname)
            if network.dhcpclass:
                f.write("DHCP_CLASSID=%s\n" % network.dhcpclass)

        if network.mtu:
            f.write("MTU=%s\n" % network.mtu)

        f.close()

    def writeNetworkKey(self, instroot, network):
        if not network.wepkey:
            return

        path = instroot + "/etc/sysconfig/network-scripts/keys-" + network.device
        f = file(path, "w+")
        os.chmod(path, 0600)
        f.write("KEY=%s\n" % network.wepkey)
        f.close()

    def writeNetworkConfig(self, instroot, useipv6, hostname, gateway):
        path = instroot + "/etc/sysconfig/network"
        f = file(path, "w+")
        os.chmod(path, 0644)

        f.write("NETWORKING=yes\n")

        if useipv6:
            f.write("NETWORKING_IPV6=yes\n")
        else:
            f.write("NETWORKING_IPV6=no\n")

        if hostname:
            f.write("HOSTNAME=%s\n" % hostname)
        else:
            f.write("HOSTNAME=localhost.localdomain\n")

        if gateway:
            f.write("GATEWAY=%s\n" % gateway)

        f.close()

    def writeNetworkHosts(self, instroot, hostname):
        localline = ""
        if hostname and hostname != "localhost.localdomain":
            localline += hostname + " "
            l = string.split(hostname, ".")
            if len(l) > 1:
                localline += l[0] + " "
        localline += "localhost.localdomain localhost"

        path = instroot + "/etc/hosts"
        f = file(path, "w+")
        os.chmod(path, 0644)
        f.write("127.0.0.1\t\t%s\n" % localline)
        f.write("::1\t\tlocalhost6.localdomain6 localhost6\n")
        f.close()

    def writeNetworkResolv(self, instroot, nodns, nameservers):
        if nodns or not nameservers:
            return

        path = instroot + "/etc/resolv.conf"
        f = file(path, "w+")
        os.chmod(path, 0644)

        for ns in (nameservers):
            if ns:
                f.write("nameserver %s\n" % ns)

        f.close()

    def configureNetwork(self):
        instroot = self.build_dir + "/install_root"

        try:
            os.makedirs(instroot + "/etc/sysconfig/network-scripts")
        except OSError, (err, msg):
            if err != errno.EEXIST:
                raise

        useipv6 = False
        nodns = False
        hostname = None
        gateway = None
        nameservers = None

        for network in self.cfg.ksobj._get("network","network"):
            if not network.device:
                raise InstallationError(_("No --device specified with network kickstart command"))

            if network.onboot and network.bootProto.lower() != "dhcp" and \
               not (network.ip and network.netmask):
                raise InstallationError(_("No IP address and/or netmask specified with static " +
                                        "configuration for '%s'") % network.device)

            self.writeNetworkIfCfg(instroot, network)
            self.writeNetworkKey(instroot, network)

            if network.ipv6:
                useipv6 = True
            if network.nodns:
                nodns = True

            if network.hostname:
                hostname = network.hostname
            if network.gateway:
                gateway = network.gateway

            if network.nameserver:
                nameservers = string.split(network.nameserver, ",")

        self.writeNetworkConfig(instroot, useipv6, hostname, gateway)
        self.writeNetworkHosts(instroot, hostname)
        self.writeNetworkResolv(instroot, nodns, nameservers)

    def configureSystem(self):
        instroot = "%s/install_root" %(self.build_dir,)

        # FIXME: this is a bit ugly, but with the current kickstart
        # API, we don't really have a lot of choice.  it'd be nice to
        # be able to do something different, but so it goes

        # set up the language
        lang = self.cfg.ksobj._get("lang","lang") or "en_US.UTF-8"
        f = open("%s/etc/sysconfig/i18n" %(instroot,), "w+")
        f.write("LANG=\"%s\"\n" %(lang,))
        f.close()

        # next, the keyboard
        # FIXME: should this impact the X keyboard config too???
        # or do we want to make X be able to do this mapping
        import rhpl.keyboard
        k = rhpl.keyboard.Keyboard()
        if self.cfg.ksobj._get("keyboard","keyboard"):
            k.set(self.cfg.ksobj._get("keyboard","keyboard"))
        k.write(instroot)

        # next up is timezone
        tz = self.cfg.ksobj._get("timezone","timezone") or "America/New_York"
        utc = self.cfg.ksobj._get("timezone","isUtc")
        f = open("%s/etc/sysconfig/clock" %(instroot,), "w+")
        f.write("ZONE=\"%s\"\n" %(tz,))
        f.write("UTC=%s\n" %(utc,))
        f.close()

        # do any authconfig bits
        # authconfig is in authconfig and may not be installed
        if os.path.exists("%s/usr/sbin/authconfig" %(instroot,)):
            auth = self.cfg.ksobj._get("authconfig","authconfig") or "--useshadow --enablemd5"
            args = ["/usr/sbin/authconfig", "--update", "--nostart"]
            args.extend(auth.split())
            subprocess.call(args, preexec_fn=self.run_in_root)
        elif os.path.exists("%s/usr/sbin/pwconv" %(instroot,)):
            if self.cfg.lm_use_shadow:
                subprocess.call(["/usr/sbin/pwconv"],preexec_fn=self.run_in_root)
                subprocess.call(["/usr/sbin/grpconv"],preexec_fn=self.run_in_root)

        # firewall.  FIXME: should handle the rest of the options
        # lokkit is system-config-securitylevel-tui, forget about it
        if self.cfg.ksobj._get("firewall","enabled"):
            if os.path.exists("%s/usr/sbin/lokkit" %(instroot,)):
                subprocess.call(["/usr/sbin/lokkit", "-f", "--quiet",
                                "--nostart", "--enabled"],
                                preexec_fn=self.run_in_root)
            elif os.path.exists("%s/sbin/chkconfig" %(instroot,)):
                subprocess.call(["/sbin/chkconfig", "iptables", "on"], preexec_fn=self.run_in_root)

        # selinux
        if os.path.exists("/selinux/enforce"):
            if os.path.exists("%s/etc/selinux/config" %(instroot,)):
                # lokkit is system-config-securitylevel-tui, may not be installed
                # sed is sed package, may not be installed
                f = open("%s/etc/selinux/config" %(instroot,), "rw+")
                buf = f.read()
                self.log.debug(_("Setting SELinux to: %s") % str(self.cfg.ksobj._get("selinux","selinux")))
                if self.cfg.ksobj._get("selinux","selinux") == kickstart.constants.SELINUX_PERMISSIVE:
                    buf = buf.replace("SELINUX=enforcing","SELINUX=permissive")
                elif self.cfg.ksobj._get("selinux","selinux") == kickstart.constants.SELINUX_DISABLED:
                    buf = buf.replace("SELINUX=enforcing","SELINUX=disabled")

                f.seek(0)
                f.write(buf)
                f.close()
        else:
            self.log.error(_("SELinux requested but not enabled on the host system."))
            self.log.info(_("Disabling SELinux as we cannot restore contexts"))
            if os.path.exists("%s/etc/selinux/config" %(instroot,)):
                # lokkit is system-config-securitylevel-tui, may not be installed
                # sed is sed package, may not be installed
                f = open("%s/etc/selinux/config" %(instroot,), "rw+")
                buf = f.read()
                buf = buf.replace("SELINUX=enforcing","SELINUX=disabled")
                f.seek(0)
                f.write(buf)
                f.close()

        # Set the root password
        if self.cfg.ksobj._get("rootpw","isCrypted"):
            self.log.debug(_("Setting crypted root password"))
            subprocess.call(["/usr/sbin/usermod", "-p", self.cfg.ksobj._get("rootpw","password"), "root"], preexec_fn=self.run_in_root)
        elif self.cfg.ksobj._get("rootpw","password") == "":
            # Root password is not set and not crypted, empty it
            self.log.debug(_("Emptying root password"))
            subprocess.call(["/usr/bin/passwd", "-d", "root"], preexec_fn=self.run_in_root)
        else:
            # Root password is set and not crypted
            self.log.debug(_("Setting uncrypted root password"))
            p1 = subprocess.Popen(["/bin/echo", self.cfg.ksobj._get("rootpw","password")], stdout=subprocess.PIPE, preexec_fn=self.run_in_root)
            p2 = subprocess.Popen(["/usr/bin/passwd", "--stdin", "root"], stdin=p1.stdout, stdout=subprocess.PIPE, preexec_fn=self.run_in_root)
            (output, error) = p2.communicate()
            self.log.debug(_("Output: %r, Error: %r") % (output,error))

        # enable/disable services appropriately
        if os.path.exists("%s/sbin/chkconfig" %(instroot,)):
            for s in self.cfg.ksobj._get("services","enabled"):
                subprocess.call(["/sbin/chkconfig", s, "on"],
                                preexec_fn=self.run_in_root)
            for s in self.cfg.ksobj._get("services","disabled"):
                subprocess.call(["/sbin/chkconfig", s, "off"],
                                preexec_fn=self.run_in_root)

    def runPost(self):
        instroot = "%s/install_root" %(self.build_dir,)
        # and now, for arbitrary %post scripts
        for s in filter(lambda s: s.type == kickstart.parser.KS_SCRIPT_POST,
                        self.cfg.ksobj._get("scripts")):
            (fd, path) = tempfile.mkstemp("", "ks-script-", "%s/tmp" %(instroot,))
            os.write(fd, s.script)
            os.close(fd)
            os.chmod(path, 0700)

            if not s.inChroot:
                env = {"BUILD_DIR": self.build_dir,
                       "INSTALL_ROOT": "%s/install_root" %(self.build_dir,),
                       "LIVE_ROOT": "%s/out" %(self.build_dir,)}
                preexec = lambda: os.chdir(self.build_dir,)
                script = path
            else:
                env = {}
                preexec = self.run_in_root
                script = "/tmp/%s" %(os.path.basename(path),)

            try:
                subprocess.call([s.interp, script],
                                preexec_fn = preexec, env = env)
            except OSError, (err, msg):
                os.unlink(path)
                raise InstallationError("Failed to execute %%post script with '%s' : %s" % (s.interp, msg))
            os.unlink(path)

    def get_kernel_version(self):
        #
        # FIXME: this doesn't handle multiple kernels - we should list
        #        them all in the isolinux menu
        #
        self.kernel_types = [
                                         {
                                           "regexp": ".*PAE-debug$",
                                           "initrd": "initpaed.img",
                                           "vmlinuz": "vmlpaed",
                                           "type": "PAE-debug"
                                         },

                                         {
                                           "regexp": ".*PAE$",
                                           "initrd": "initpae.img",
                                           "vmlinuz": "vmlpae",
                                           "type": "PAE"
                                         },

                                         {
                                            "regexp": ".*xen$",
                                            "initrd": "initxen.img",
                                            "vmlinuz": "vmlxen",
                                            "type": "xen"
                                         },

                                         {
                                            "regexp": ".*debug$",
                                            "initrd": "initdbg.img",
                                            "vmlinuz": "vmldebug",
                                            "type": "debug"
                                         },

                                         {
                                            "regexp": "^.*$",
                                            "initrd": "initrd.img",
                                            "vmlinuz": "vmlinuz",
                                            "type": "normal"
                                         }
                            ]

        kernel_dirs = []
        kernels = []
        modules_dir = "%s/install_root/lib/modules" % self.build_dir

        if os.path.isdir(modules_dir):
            kernel_dirs = os.listdir(modules_dir)

        kernel_dirs.sort()

        if not kernel_dirs:
            raise InstallationError(_("No kernels installed: /lib/modules is empty"))

        for kernel_dir in kernel_dirs:
            found = False
            for kernel_type in self.kernel_types:
                if not found:
                    if re.match(kernel_type["regexp"], kernel_dir):
                        kernels.append({"name": kernel_dir, "initrd": kernel_type["initrd"], "vmlinuz": kernel_type["vmlinuz"], "type": kernel_type["type"] })
                        found = True

        self.log.debug("Kernels: %s" % str(kernels))
        return kernels

    def createInitramfs(self):
        # Create initramfs
        if not os.path.isfile("/usr/lib/livecd-creator/mayflower"):
            raise InstallationError(_("livecd-creator not correctly installed : " +
                                    "/usr/lib/livecd-creator/mayflower not found"))
        shutil.copy("/usr/lib/livecd-creator/mayflower",
                        "%s/install_root/sbin/mayflower" %(self.build_dir,))
        # modules needed for booting (this is butt ugly and we need to retrieve this from elsewhere, e.g. the kernel)
        mayflowerconf = open(self.build_dir + "/install_root/etc/mayflower.conf", "w")
        mayflowerconf.write('MODULES+="cdrom ide-cd ahci loop dm_snapshot squashfs ext3 ext2 ehci_hcd uhci_hcd ohci_hcd usb_storage sd_mod sr_mod usbhid ata_piix vfat msdos "\n')
        mayflowerconf.write('MODULES+="sata_mv sata_qstor sata_sis sata_uli sata_inic162x "\n')
        mayflowerconf.write('MODULES+="sata_nv sata_sil24 sata_svw sata_via "\n')
        mayflowerconf.write('MODULES+="sata_promise sata_sil sata_sx4 sata_vsc "\n')
        mayflowerconf.write('MODULES+="ata_generic pata_ali pata_amd pata_artop pata_atiixp pata_cmd640 pata_cmd64x pata_cs5520 pata_cs5530 pata_cs5535 pata_cypress pata_efar pata_hpt366 pata_hpt37x pata_hpt3x2n pata_hpt3x3 pata_isapnp pata_it8213 pata_it821x pata_jmicron pata_marvell pata_mpiix pata_netcell pata_ns87410 pata_oldpiix pata_optidma pata_opti pata_pcmcia pata_pdc2027x pata_pdc202xx_old pata_qdi pata_serverworks pata_sil680 pata_sis pata_sl82c105 pata_triflex pata_via pdc_adma "\n')
        mayflowerconf.write('MODULES+="sym53c8xx aic7xxx "\n')
        mayflowerconf.close()

        for kernel in self.get_kernel_version():
            subprocess.call(["/sbin/mayflower", "-f", "/boot/livecd-initramfs-%s.img" % kernel["name"],
                            kernel["name"]],
                            preexec_fn=self.run_in_root)

        for f in ("/sbin/mayflower", "/etc/mayflower.conf"):
            os.unlink("%s/install_root/%s" %(self.build_dir, f))

    def relabelSystem(self):
        # finally relabel all files
        if self.cfg.ksobj._get("selinux","selinux"):
            instroot = "%s/install_root" %(self.build_dir,)
            if os.path.exists("%s/sbin/restorecon" %(instroot,)):
                subprocess.call(["/sbin/restorecon", "-v", "-r", "/"],
                                preexec_fn=self.run_in_root)

    def prelinkSystem(self):
        # prelink the system
        instroot = "%s/install_root" %(self.build_dir,)
        if os.path.exists("%s/usr/sbin/prelink" %(instroot,)):
            subprocess.call(["/usr/sbin/prelink", "-mRaN"],
                            preexec_fn=self.run_in_root)
        return True

    def launchShell(self):
        subprocess.call(["/bin/bash"], preexec_fn=self.run_in_root)

    def configureBootloader(self):
        """configure the boot loader"""

        # set up boot loader
        #
        # TODO:
        #  - fix for archs not using grub
        #  - fix for non-i386
        #  - error handling
        #

        # Handle multiple kernels
        for kernel in self.get_kernel_version():
            # Copy appr. vmlinuz in place
            shutil.copyfile("%s/install_root/boot/vmlinuz-%s"
                            %(self.build_dir, kernel["name"]),
                            "%s/out/isolinux/%s" %(self.build_dir, kernel["vmlinuz"]))

            # Copy appr. initrd in place
            shutil.copyfile("%s/install_root/boot/livecd-initramfs-%s.img"
                            %(self.build_dir,kernel["name"]),
                            "%s/out/isolinux/%s" %(self.build_dir,kernel["initrd"]))

            # Lose whatever was there
            os.unlink("%s/install_root/boot/livecd-initramfs-%s.img"
                    %(self.build_dir,kernel["name"]))

        syslinuxfiles = ["isolinux.bin"]
        menus = ["vesamenu.c32", "menu.c32"]
        syslinuxMenu = None

        for m in menus:
            path = "%s/install_root/usr/lib/syslinux/%s" % (self.build_dir, m)
            if os.path.isfile(path):
                syslinuxfiles.append(m)
                syslinuxMenu=m
                break
        if syslinuxMenu is None:
            raise InstallationError(_("syslinux not installed : no suitable *menu.c32 found"))

        for p in syslinuxfiles:
            path = "%s/install_root/usr/lib/syslinux/%s" % (self.build_dir, p)
            if not os.path.isfile(path):
                raise InstallationError(_("syslinux not installed : %s not found") % path)

            shutil.copy(path, "%s/out/isolinux/%s" % (self.build_dir, p))

        if os.path.exists("%s/install_root/usr/lib/anaconda-runtime/syslinux-vesa-splash.jpg" %(self.build_dir,)):
            shutil.copy("%s/install_root/usr/lib/anaconda-runtime/syslinux-vesa-splash.jpg" %(self.build_dir,),
                        "%s/out/isolinux/splash.jpg" %(self.build_dir,))
            have_background = "menu background splash.jpg"
        else:
            have_background = ""

        cfg_template = """
default %(menu)s
timeout %(bootloader_timeout)s
%(bootloader_isolinux_extras)s

%(background)s
menu title """ + _("Welcome to %(label)s!") + """
menu color border 0 #ffffffff #00000000
menu color sel 7 #ffffffff #ff000000
menu color title 0 #ffffffff #00000000
menu color tabmsg 0 #ffffffff #00000000
menu color unsel 0 #ffffffff #00000000
menu color hotsel 0 #ff000000 #ffffffff
menu color hotkey 7 #ffffffff #ff000000
"""

        cfg = cfg_template % {"menu" : syslinuxMenu, "background" : have_background, "label" : self.fs_label,
                              "bootloader_timeout" : self.cfg.lm_bootloader_timeout,
                              "bootloader_isolinux_extras" : self.cfg.lm_bootloader_isolinux_extras }

# FIXME
# Change initrd.img to the "default" kernel, which at this moment we don't have

        for kernel in self.get_kernel_version():
            # Match against self.preferred_kernel_type
            if kernel["type"] == self.cfg.lm_preferred_kernel_type:
                preferred_kernel = kernel

        # short, long, extra, kernel_initrd, kernel_vmlinuz
        stanzas = [(short, long, options, preferred_kernel["initrd"], preferred_kernel["vmlinuz"]) for short, long, options in self.cfg.lm_bootloader_stanzas]

        # List all kernels
        for kernel in self.get_kernel_version():
            if not kernel["type"] == self.cfg.lm_preferred_kernel_type:
                stanzas.append( ("linux-%s" % kernel["name"], _("Run kernel %s") % kernel["name"], self.lm_bootloader_options, "%s" % kernel["initrd"], "%s" % kernel["vmlinuz"]) )

        if os.path.exists("%s/install_root/usr/lib/anaconda-runtime/checkisomd5" %(self.build_dir,)):
            stanzas.append( ("check", _("Verify and run from image"), "check", "", "") )

        bootcfg = ""

        for (short, long, extra, kernel_initrd, kernel_vmlinuz) in stanzas:
            bootcfg += """label %(short)s
  menu label %(long)s
  kernel %(kernel_vmlinuz)s
  append initrd=%(kernel_initrd)s ro root=CDLABEL=%(label)s rootfstype=iso9660 liveimg %(extra)s
"""

            bootcfg = bootcfg % {"label": self.fs_label, "background" : have_background, "short": short, "long": long, "extra": extra, "kernel_vmlinuz": kernel_vmlinuz, "kernel_initrd": kernel_initrd }

        memtest = glob.glob("%s/install_root/boot/memtest86*" %(self.build_dir,))
        if len(memtest) > 0:
            shutil.copy(memtest[0], "%s/out/isolinux/memtest" %(self.build_dir,))
            bootcfg += """label memtest
  menu label """ + _("Memory Test") + """
  kernel memtest"""

        finalcfg = cfg + bootcfg

        cfgf = open("%s/out/isolinux/isolinux.cfg" %(self.build_dir,), "w")
        cfgf.write(finalcfg)
        cfgf.close()

        # TODO: enable external entitity to partipate in adding boot entries

    def createIso(self):
        """write out the live CD ISO"""
        subprocess.call(["/usr/bin/mkisofs", "-o", "%s.iso" %(self.fs_label,),
                         "-b", "isolinux/isolinux.bin",
                         "-c", "isolinux/boot.cat",
                         "-no-emul-boot", "-boot-load-size", "4",
                         "-boot-info-table", "-J", "-r", "-hide-rr-moved",
                         "-V", "%s" %(self.fs_label,), "%s/out" %(self.build_dir)])

        # implant an isomd5sum
        if os.path.exists("/usr/lib/anaconda-runtime/implantisomd5"):
            subprocess.call(["/usr/lib/anaconda-runtime/implantisomd5",
                             "%s.iso" %(self.fs_label,)])
        else:
            print >> sys.stderr, _("anaconda-runtime not installed; not setting up mediacheck")

    def createSquashFS(self):
        """create compressed squashfs file system"""
        if not self.skip_compression:
            # FIXME: mksquashfs segfaults if PWD isn't set in the environment
            subprocess.call(["/sbin/mksquashfs", "os.img", "sysroot",
                             "../out/squashfs.img"],
                            cwd="%s/data" %(self.build_dir,),
                            env={"PWD": "%s/data" %(self.build_dir,)})
        else:
            shutil.move("%s/data/os.img" %(self.build_dir,),
                        "%s/out/ext3fs.img" %(self.build_dir,))

    #
    # cleanupDeleted removes unused data from the sparse ext3 os image file.
    # The process involves: resize2fs-to-minimal, truncation,
    # resize2fs-to-uncompressed-size (with implicit resparsification)
    #
    def cleanupDeleted(self):
        # e2fsck -f -y is required by resize2fs
        subprocess.call(["/sbin/e2fsck", "-f", "-y", "os.img"],
                        cwd="%s/data" %(self.build_dir,),
                        env={"PWD": "%s/data" %(self.build_dir,)})


        (xxmode,xxino,xxdev,xxnlink,xxuid,xxgid,xxsize,xxatime,
         xxmtime,xxctime) = os.stat(self.build_dir + "/data/os.img")
        kbsize = xxsize / 1024L

        # resize2fs doesn't have any kind of minimal setting, so use
        # a binary search to get it to minimal size.
        # TODO: if a --verbose/debug flag exists, don't devnull the output
        FNULL = os.open('/dev/null', os.O_WRONLY)
        size_top = kbsize / 4L
        size_bot = 0
        while size_top != (size_bot + 1):
            trysize = (size_bot + ((size_top - size_bot) / 2)) * 4
            resize2fs_retval = subprocess.call(["/sbin/resize2fs", "os.img",
                                                "%dK" %(trysize,)],
                                               cwd="%s/data" %(self.build_dir,),
                                               env={"PWD": "%s/data" %(self.build_dir,)},
                                               stdout=FNULL,
                                               stderr=FNULL)

            if not resize2fs_retval:
                size_top = (size_bot + ((size_top - size_bot) / 2))
            else:
                size_bot = (size_bot + ((size_top - size_bot) / 2))


        os.close(FNULL)

        print >> sys.stderr, _("Installation target minimized to %dK") % (size_top * 4)

        # truncate the unused excess portion of the sparse file
        fd = os.open("%s/data/os.img" %(self.build_dir,), os.O_WRONLY )
        os.ftruncate(fd, size_top * 4096L)
        os.close(fd)

        # resize back to uncompressed-size
        resize2fs_retval = subprocess.call(["/sbin/resize2fs", "os.img",
                                            "%dK" %(kbsize,)],
                                           cwd="%s/data" %(self.build_dir,),
                                           env={"PWD": "%s/data" %(self.build_dir,)})

