#
# Copyright 2007 Fedora Unity
#
# Jonathan Steffan <jon a fedoraunity.org>
# Jeroen van Meeuwen <kanarip a fedoraunity.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 only
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import glob
import logging
import os
import re
import shutil
import subprocess
import sys
sys.path.append('/usr/lib/anaconda-runtime')
import splittree

# Translation
from rhpl.translate import _, N_, getDefaultLangs
import rhpl.translate as translate

class RevisorPungi:
    def __init__(self, base):
        self.base = base
        self.cfg = base.cfg
        self.log = base.log

        self.destdir = os.path.join(self.cfg.working_directory,"revisor-pungi")
        self.archdir = os.path.join(self.destdir,
                                   self.cfg.version,
                                   self.cfg.revisor_model,
                                   self.cfg.architecture)

        self.topdir = os.path.join(self.archdir, 'os')
        self.isodir = os.path.join(self.archdir, 'iso')

        self.workdir = os.path.join(self.destdir,
                                    'work',
                                    self.cfg.revisor_model,
                                    self.cfg.architecture)

        self.common_files = []
        self.infofile = os.path.join(self.destdir,
                                     '.composeinfo')

        if not os.path.exists(self.destdir):
            try:
                os.makedirs(self.destdir)
            except OSError, e:
                self.log.error(_("Error: Cannot create destination dir %s") % self.destdir)
        else:
            # Empty it
            try:
                shutil.rmtree(self.destdir)
            except OSError, e:
                self.log.error(_("Error: Cannot remove destination dir %s") % self.destdir)
            # Recreate it
            try:
                os.makedirs(self.destdir)
            except OSError, e:
                self.log.error(_("Error: Cannot recreate destination dir %s") % self.destdir)

    def doCreateRepo(self, database=True, callback=None):

        # create repodata for the tree
        createrepo = ['/usr/bin/createrepo']
        createrepo.append('-v')

        if database:
            createrepo.append('--database')

        createrepo.append('--groupfile')
        createrepo.append(self.cfg.comps)

        createrepo.append(self.topdir)

        # run the command
        self.base.run_command(createrepo, rundir=self.topdir, callback=callback)

    def doBuildinstall(self, callback=None):
        """Run anaconda-runtime's buildinstall on the tree."""

        # setup the buildinstall call
        buildinstall = ['/usr/lib/anaconda-runtime/buildinstall']
        #buildinstall.append('TMPDIR=%s' % self.workdir) # TMPDIR broken in buildinstall

        buildinstall.append('--product')
        buildinstall.append(self.cfg.product_name)

        if not self.cfg.revisor_model == "":
            buildinstall.append('--variant')
            buildinstall.append(self.cfg.revisor_model)

        buildinstall.append('--version')
        buildinstall.append(self.cfg.version)

        buildinstall.append('--release')
        buildinstall.append('"%s %s"' % (self.cfg.product_name, self.cfg.version))

        buildinstall.append('--prodpath')
        buildinstall.append(self.cfg.product_path)

        if hasattr(self.cfg,'bugurl'):
            buildinstall.append('--bugurl')
            buildinstall.append(self.cfg.bugurl)

        buildinstall.append(self.topdir)

        # run the command
        self.base.run_command(buildinstall, callback=callback)

    def doPackageorder(self, callback=None):
        """Run anaconda-runtime's pkgorder on the tree, used for splitting media."""
        pkgorderfile = open(os.path.join(self.workdir, 'pkgorder-%s' % self.cfg.architecture), 'w')
        # setup the command
        pkgorder = ['/usr/lib/anaconda-runtime/pkgorder', '--debug']
        #pkgorder.append('TMPDIR=%s' % self.workdir)
        pkgorder.append(self.topdir)
        pkgorder.append(self.cfg.architecture)
        pkgorder.append(self.cfg.product_path)

        # run the command
        self.base.run_command(pkgorder, output=pkgorderfile, callback=callback)
        pkgorderfile.close()

    def doGetRelnotes(self, callback=None):
        """Get extra files from packages in the tree to put in the topdir of
           the tree."""

        self.log.debug(_("Getting relnotes..."))

        docsdir = os.path.join(self.workdir, 'docs')
        os.makedirs(docsdir)

        # Expload the packages we list as relnote packages
        pkgs = os.listdir(os.path.join(self.topdir, self.cfg.product_path))

        self.log.debug(str(pkgs))
        rpm2cpio = ['/usr/bin/rpm2cpio']
        cpio = ['cpio', '-imud']

        for pkg in pkgs:
            for pattern in self.cfg.release_pkgs.split():
                if re.match(pattern, pkg):
                    extraargs = [os.path.join(self.topdir, self.cfg.product_path, pkg)]
                    try:
                        p1 = subprocess.Popen(rpm2cpio + extraargs, cwd=docsdir, stdout=subprocess.PIPE)
                        (out, err) = subprocess.Popen(cpio, cwd=docsdir, stdin=p1.stdout, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, universal_newlines=True).communicate()
                    except:
                        self.log.error(_("An error occured extracting the release files: %s") % err)

                    self.log.debug(out)

        # Walk the tree for our files
        for dirpath, dirname, filelist in os.walk(docsdir):
            for directory in dirname:
                for pattern in self.cfg.release_files.split():
                    if re.match(pattern,directory) and not os.path.exists(os.path.join(self.topdir, directory)):
                        self.log.debug(_("Copying %s") % directory)
                        shutil.copytree(os.path.join(dirpath, directory), os.path.join(self.topdir, directory))

            for filename in filelist:
                for pattern in self.cfg.release_files.split():
                    if re.match(pattern,filename) and not os.path.exists(os.path.join(self.topdir, filename)):
                        self.log.debug("Copying release note file %s" % filename)
                        shutil.copy(os.path.join(dirpath, filename), os.path.join(self.topdir, filename))
                        self.common_files.append(filename)

        self.log.debug(str(self.common_files))

    def doCopyDir(self, copy_dir=None):
        if not copy_dir:
            return

        if not os.access(self.cfg.copy_dir, os.R_OK):
            self.log.error(_("Could not access directory %s, cannot copy extra files onto the media.") % self.cfg.copy_dir)
            return

        try:
            os.makedirs(os.path.join(self.topdir,"files"))
        except:
            self.log.error(_("Could not create files/ directory: %s") % os.path.join(self.topdir,"files"))

        if not self.cfg.copy_dir == "":
            if os.access(self.cfg.copy_dir, os.R_OK):
                # So what is the destination directory?
                src_dir = self.cfg.copy_dir
                dst_dir = "%s/files/%s"

                for root, dirs, files in os.walk(src_dir):
                    real_dst = dst_dir % (self.topdir, root.replace(src_dir,''))
                    if not os.access(real_dst, os.R_OK):
                        self.log.debug(_("Creating %s") % real_dst)
                        os.makedirs(real_dst)

                    for file in files:
                        self.log.debug(_("Copying %s to %s") % (os.path.join(root,file), os.path.join(real_dst,file)))
                        shutil.copyfile(os.path.join(root,file), os.path.join(real_dst,file))

            else:
                self.log.warning(_("copy_dir '%s' not accessible") % self.cfg.copy_dir)

    def doSplittree(self, media_size = 0, discdir = "unified", callback=None):
        """Use anaconda-runtime's splittree to split the tree into appropriate
        sized chunks."""

        # Hey I get size in bytes but I want MBytes
        timber = splittree.Timber()
        timber.arch = self.cfg.architecture
        timber.target_size = float(media_size)
        timber.total_discs = int(self.cfg.mediatypes[discdir]["discs"])
        timber.bin_discs = int(self.cfg.mediatypes[discdir]["discs"])
        timber.src_discs = 0
        timber.release_str = '%s %s' % (self.cfg.product_name, self.cfg.version)
        timber.package_order_file = self.cfg.pkgorder_file
        timber.dist_dir = "%s-%s" % (self.topdir,discdir)
        self.log.debug("dist_dir is %s" % timber.dist_dir)
        timber.src_dir = os.path.join(self.destdir, self.cfg.version, self.cfg.revisor_model, 'source', 'SRPMS')
        self.log.debug("src_dir is %s" % timber.src_dir)
        timber.product_path = self.cfg.product_path
        timber.common_files = self.common_files
        #timber.reserve_size =

        output = timber.main()
        del timber

    def doCreateSplitrepo(self, discdir = "unified", callback = None):
        """Create the split metadata for the isos"""

        if int(self.cfg.mediatypes[discdir]["discs"]) > 1:
            discinfo = open('%s-%s-disc1/.discinfo' % (self.topdir,discdir), 'r').readlines()
        else:
            discinfo = open(os.path.join(self.topdir, '.discinfo'), 'r').readlines()

        mediaid = discinfo[0].rstrip('\n')

        # set up the process
        createrepo = ['/usr/bin/createrepo']
        createrepo.append('-v')
        createrepo.append('--database')

        createrepo.append('--groupfile')
        createrepo.append(self.cfg.comps)

        createrepo.append('--baseurl')
        createrepo.append('media://%s' % mediaid)

        createrepo.append('--outputdir')
        if int(self.cfg.mediatypes[discdir]["discs"]) == 1:
            os.makedirs('%s-%s-disc1' % (self.topdir,discdir))
        createrepo.append('%s-%s-disc1' % (self.topdir,discdir))

        createrepo.append('--basedir')
        if int(self.cfg.mediatypes[discdir]["discs"]) == 1:
            createrepo.append(self.topdir)
            createrepo.append(self.topdir)
        else:
            createrepo.append('%s-%s-disc1' % (self.topdir,discdir))

        if int(self.cfg.mediatypes[discdir]["discs"]) > 1:
            createrepo.append('--split')

            for disc in range(1, int(self.cfg.mediatypes[discdir]["discs"]) + 1):
                createrepo.append('%s-%s-disc%s' % (self.topdir, discdir, disc))

        # run the command
        self.base.run_command(createrepo, callback=callback)

    def doCreateIso(self, mediatype=None, disc=0,  callback=None):

        mt = self.cfg.mediatypes[mediatype]
        # Based on what we have, create some names and stuff
        isoname = "%s-%s-%s" % (self.cfg.iso_basename, self.cfg.version, self.cfg.architecture)
        if len(mt["label"]) > 1:
            isoname = "%s-%s" % (isoname, mt["label"])
        if mt["discs"] > 1:
            isoname = "%s%d" % (isoname, disc)

        isoname = "%s.iso" % isoname

        isofile = os.path.join(self.cfg.destination_directory,"iso",isoname)

        extraargs = []
        if disc == 1 or disc == 0:
            if self.cfg.architecture == 'i386' or self.cfg.architecture == 'x86_64':
                extraargs.extend(self.cfg.x86bootargs)
            elif self.cfg.architecture == 'ia64':
                extraargs.extend(self.cfg.ia64bootargs)
            elif self.cfg.architecture == 'ppc':
                extraargs.extend(self.cfg.ppcbootargs)
                extraargs.append(os.path.join('%s' % (self.topdir), "ppc/mac"))

        extraargs.append('-V')
        volume = "%s %s %s" % (self.cfg.iso_label, self.cfg.version, self.cfg.architecture)
        if mt["label"] > 1:
            volume = "%s %s" % (volume, mt["label"])
        if mt["discs"] > 1:
            volume = "%s%d" % (volume, disc)

        extraargs.extend(['"%s"' % volume, '-o', isofile])

        if disc == 0:
            extraargs.append(self.topdir)
            num_packages = len(os.listdir(os.path.join(self.topdir, self.cfg.product_path)))
        else:
            extraargs.append("%s-%s-disc%s" % (self.topdir, mt["discdir"], disc))
            num_packages = len(os.listdir(os.path.join('%s-%s-disc%s/' % (self.topdir, mt["discdir"], disc), self.cfg.product_path)))

        self.cfg.built_iso_images.append({'type': mediatype, 'location': isofile, 'packages': num_packages})

        self.base.run_command(self.cfg.cmd_mkisofs + extraargs, callback=callback)

    def create_rescue_disk(self):
        # Now make rescue images
        if not self.cfg.architecture == 'source' and \
            os.path.exists('/usr/lib/anaconda-runtime/mk-rescueimage.%s' % self.cfg.architecture):
            isoname = '%s-%s-%s-rescuecd.iso' % (self.cfg.iso_basename,
                self.cfg.version, self.cfg.architecture)
            isofile = os.path.join(self.cfg.destination_directory,"iso",isoname)

            # make the rescue tree
            rescue = ['/usr/lib/anaconda-runtime/mk-rescueimage.%s' % self.cfg.architecture]
            rescue.append(self.topdir)
            rescue.append(self.workdir)
            rescue.append(self.cfg.iso_basename)
            rescue.append(self.cfg.product_path)

            # run the command
            self.base.run_command(rescue)

            # write the iso
            extraargs = []

            if self.cfg.architecture == 'i386' or self.cfg.architecture == 'x86_64':
                extraargs.extend(x86bootargs)
            elif self.cfg.architecture == 'ia64':
                extraargs.extend(ia64bootargs)
            elif self.cfg.architecture == 'ppc':
                extraargs.extend(ppcbootargs)
                extraargs.append(os.path.join(self.workdir, "%s-rescueimage" % self.cfg.architecture, "ppc/mac"))

            extraargs.append('-V')
            extraargs.append('"%s %s %s Rescue"' % (self.cfg.product_name,
                    self.cfg.version, self.cfg.architecture))

            extraargs.append('-o')
            extraargs.append(os.path.join(self.cfg.destination_directory,"iso",isofile))

            extraargs.append(os.path.join(self.workdir, "%s-rescueimage" % self.cfg.architecture))

            # run the command
            self.base.run_command(mkisofs + extraargs, callback=callback)

            # implant md5 for mediacheck on all but source arches
            if not self.cfg.architecture == 'source':
                self.base.run_command(['/usr/lib/anaconda-runtime/implantisomd5', isofile])

            # shove the sha1sum into a file
            sha1file = open(os.path.join(self.cfg.destination_directory,"iso","SHA1SUM"), 'a')
            self.base.run_command(['/usr/bin/sha1sum', isoname], rundir=os.path.join(self.cfg.destination_directory,"iso"), output=sha1file)
            sha1file.close()

    def cleanup(self):
# FIXME
        # Do some clean up
        dirs = os.listdir(self.archdir)

        for directory in dirs:
            if directory.startswith('os-disc') or directory.startswith('SRPM-disc'):
                shutil.move(os.path.join(self.archdir, directory), os.path.join(self.workdir, directory))

